"""Container startup orchestration for vLLM Semantic Router."""

import os
import subprocess

from cli.config_generator import generate_envoy_config_from_user_config
from cli.consts import (
    DEFAULT_NOFILE_LIMIT,
    MIN_NOFILE_LIMIT,
    PLATFORM_AMD,
)
from cli.docker_images import (
    _normalize_platform,
    get_docker_image,
    get_runtime_images,
)
from cli.docker_openclaw_support import configure_openclaw_support
from cli.docker_run_command import (
    append_env_vars,
    append_host_gateway,
    append_mount_specs,
    append_port_mappings,
    build_base_run_command,
    maybe_append_amd_gpu_passthrough,
)
from cli.docker_runtime import get_container_runtime, resolve_docker_cli_path
from cli.docker_services import (
    docker_container_status,
    docker_remove_container,
    docker_stop_container,
)
from cli.parser import parse_user_config
from cli.runtime_stack import RuntimeStackLayout, resolve_runtime_stack
from cli.runtime_topology import resolve_runtime_topology, split_runtime_enabled
from cli.utils import get_logger

log = get_logger(__name__)


def docker_start_vllm_sr(
    config_file,
    env_vars,
    listeners,
    image=None,
    router_image=None,
    envoy_image=None,
    dashboard_image=None,
    topology=None,
    pull_policy=None,
    network_name=None,
    openclaw_network_name=None,
    minimal=False,
    stack_layout: RuntimeStackLayout | None = None,
    state_root_dir: str | None = None,
    runtime_config_file: str | None = None,
):
    """
    Start vLLM Semantic Router container.

    Returns:
        (return_code, stdout, stderr)
    """
    runtime = get_container_runtime()
    env_vars = dict(env_vars or {})
    stack_layout = stack_layout or resolve_runtime_stack()
    runtime_topology = resolve_runtime_topology(topology)
    use_split_runtime = split_runtime_enabled(runtime_topology)

    normalized_platform = _resolve_platform(env_vars)
    nofile_limit = _resolve_nofile_limit()

    runtime_network_name = (
        network_name or openclaw_network_name or stack_layout.network_name
    )
    config_dir, runtime_paths, runtime_container_config = _prepare_runtime_paths(
        config_file,
        runtime_config_file=runtime_config_file,
        state_root_dir=state_root_dir,
    )
    common_env = _build_common_runtime_env(
        env_vars,
        stack_layout,
        runtime_container_config=runtime_container_config,
        split_runtime=use_split_runtime,
    )
    if use_split_runtime:
        _render_split_envoy_config(
            runtime_paths["effective_config_path"],
            runtime_paths["envoy_config_path"],
            stack_layout,
        )
    container_specs = _resolve_container_specs(
        runtime=runtime,
        image=image,
        router_image=router_image,
        envoy_image=envoy_image,
        dashboard_image=dashboard_image,
        pull_policy=pull_policy,
        normalized_platform=normalized_platform,
        nofile_limit=nofile_limit,
        runtime_network_name=runtime_network_name,
        common_env=common_env,
        listeners=listeners,
        minimal=minimal,
        config_dir=config_dir,
        runtime_paths=runtime_paths,
        openclaw_network_name=openclaw_network_name,
        stack_layout=stack_layout,
        use_split_runtime=use_split_runtime,
    )

    log.info(f"Starting vLLM Semantic Router runtime with {runtime}...")
    started_containers = []
    stdout_chunks = []
    stderr_chunks = []

    for service_name, container_name, cmd in container_specs:
        log.info(f"Starting {service_name} container: {container_name}")
        log.debug(f"Container command: {' '.join(cmd)}")
        try:
            result = subprocess.run(cmd, capture_output=True, text=True, check=True)
        except subprocess.CalledProcessError as exc:
            if exc.stdout:
                stdout_chunks.append(exc.stdout)
            stderr_chunks.append(exc.stderr)
            _cleanup_started_containers(started_containers)
            return (exc.returncode, "\n".join(stdout_chunks), "\n".join(stderr_chunks))

        started_containers.append(container_name)
        if result.stdout:
            stdout_chunks.append(result.stdout)
        if result.stderr:
            stderr_chunks.append(result.stderr)

    return (0, "\n".join(stdout_chunks), "\n".join(stderr_chunks))


def _build_common_runtime_env(
    env_vars: dict[str, str],
    stack_layout: RuntimeStackLayout,
    *,
    runtime_container_config: str | None,
    split_runtime: bool,
):
    common_env = dict(env_vars or {})
    if runtime_container_config:
        common_env.setdefault("VLLM_SR_RUNTIME_CONFIG_PATH", runtime_container_config)
    if split_runtime:
        common_env.setdefault(
            "VLLM_SR_ROUTER_CONTAINER_NAME", stack_layout.router_container_name
        )
        common_env.setdefault(
            "VLLM_SR_ENVOY_CONTAINER_NAME", stack_layout.envoy_container_name
        )
        common_env.setdefault(
            "VLLM_SR_DASHBOARD_CONTAINER_NAME", stack_layout.dashboard_container_name
        )
    else:
        _strip_split_runtime_envs(common_env)
    return common_env


def _resolve_container_specs(
    *,
    runtime: str,
    image: str | None,
    router_image: str | None,
    envoy_image: str | None,
    dashboard_image: str | None,
    pull_policy: str | None,
    normalized_platform: str,
    nofile_limit: int,
    runtime_network_name: str,
    common_env: dict[str, str],
    listeners,
    minimal: bool,
    config_dir: str,
    runtime_paths: dict[str, str],
    openclaw_network_name: str | None,
    stack_layout: RuntimeStackLayout,
    use_split_runtime: bool,
):
    if use_split_runtime:
        runtime_images = get_runtime_images(
            image=image,
            router_image=router_image,
            envoy_image=envoy_image,
            dashboard_image=dashboard_image,
            pull_policy=pull_policy,
            platform=normalized_platform,
        )
        return _runtime_container_specs(
            runtime=runtime,
            image_by_service=runtime_images,
            nofile_limit=nofile_limit,
            runtime_network_name=runtime_network_name,
            normalized_platform=normalized_platform,
            common_env=common_env,
            listeners=listeners,
            minimal=minimal,
            config_dir=config_dir,
            runtime_paths=runtime_paths,
            openclaw_network_name=openclaw_network_name,
            stack_layout=stack_layout,
        )

    if any(value for value in (router_image, envoy_image, dashboard_image)):
        log.warning(
            "Ignoring router/envoy/dashboard image overrides for legacy topology"
        )

    selected_image = get_docker_image(
        image=image,
        pull_policy=pull_policy,
        platform=normalized_platform,
    )
    return [
        (
            "runtime",
            stack_layout.container_name,
            _build_legacy_runtime_command(
                runtime=runtime,
                image=selected_image,
                container_name=stack_layout.container_name,
                nofile_limit=nofile_limit,
                network_name=runtime_network_name,
                env_vars=common_env,
                mount_specs=_legacy_runtime_mount_specs(
                    runtime_paths,
                    include_dashboard_data=not minimal,
                ),
                listeners=listeners,
                minimal=minimal,
                enable_amd_gpu=normalized_platform == PLATFORM_AMD,
                config_dir=config_dir,
                openclaw_network_name=openclaw_network_name,
                stack_layout=stack_layout,
            ),
        )
    ]


def _strip_split_runtime_envs(env_vars: dict[str, str]) -> None:
    for key in (
        "VLLM_SR_ROUTER_CONTAINER_NAME",
        "VLLM_SR_ENVOY_CONTAINER_NAME",
        "VLLM_SR_DASHBOARD_CONTAINER_NAME",
        "TARGET_ROUTER_API_URL",
        "TARGET_ROUTER_METRICS_URL",
        "TARGET_ENVOY_URL",
        "TARGET_ENVOY_ADMIN_URL",
        "OPENCLAW_DASHBOARD_CONTAINER_NAME",
        "OPENCLAW_MODEL_GATEWAY_CONTAINER_NAME",
    ):
        env_vars.pop(key, None)


def _runtime_container_specs(
    *,
    runtime: str,
    image_by_service: dict[str, str],
    nofile_limit: int,
    runtime_network_name: str,
    normalized_platform: str,
    common_env: dict[str, str],
    listeners,
    minimal: bool,
    config_dir: str,
    runtime_paths: dict[str, str],
    openclaw_network_name: str | None,
    stack_layout: RuntimeStackLayout,
):
    listener_port = _primary_listener_port(listeners)
    setup_mode = str(common_env.get("VLLM_SR_SETUP_MODE", "")).lower() == "true"
    router_cmd = _build_router_runtime_command(
        runtime=runtime,
        router_image=image_by_service["router"],
        nofile_limit=nofile_limit,
        runtime_network_name=runtime_network_name,
        normalized_platform=normalized_platform,
        common_env=common_env,
        runtime_paths=runtime_paths,
        setup_mode=setup_mode,
        stack_layout=stack_layout,
    )
    envoy_cmd = _build_envoy_runtime_command(
        runtime=runtime,
        envoy_image=image_by_service["envoy"],
        nofile_limit=nofile_limit,
        runtime_network_name=runtime_network_name,
        common_env=common_env,
        listeners=listeners,
        runtime_paths=runtime_paths,
        setup_mode=setup_mode,
        stack_layout=stack_layout,
    )

    specs = [
        ("router", stack_layout.router_container_name, router_cmd),
        ("envoy", stack_layout.envoy_container_name, envoy_cmd),
    ]

    if minimal:
        return specs

    dashboard_cmd = _build_dashboard_runtime_command(
        runtime=runtime,
        dashboard_image=image_by_service["dashboard"],
        nofile_limit=nofile_limit,
        runtime_network_name=runtime_network_name,
        common_env=common_env,
        config_dir=config_dir,
        listener_port=listener_port,
        openclaw_network_name=openclaw_network_name,
        runtime_paths=runtime_paths,
        stack_layout=stack_layout,
    )
    specs.append(("dashboard", stack_layout.dashboard_container_name, dashboard_cmd))

    return specs


def _legacy_runtime_mount_specs(
    runtime_paths: dict[str, str],
    *,
    include_dashboard_data: bool,
):
    return _runtime_mount_specs(
        runtime_paths,
        include_models=True,
        include_dashboard_data=include_dashboard_data,
    )


def _build_legacy_runtime_command(
    *,
    runtime: str,
    image: str,
    container_name: str,
    nofile_limit: int,
    network_name: str,
    env_vars: dict[str, str],
    mount_specs: list[str],
    listeners,
    minimal: bool,
    enable_amd_gpu: bool,
    config_dir: str,
    openclaw_network_name: str | None,
    stack_layout: RuntimeStackLayout,
):
    legacy_env = _build_legacy_runtime_env(
        env_vars=env_vars,
        minimal=minimal,
        mount_specs=mount_specs,
        config_dir=config_dir,
        openclaw_network_name=openclaw_network_name,
        runtime=runtime,
        stack_layout=stack_layout,
    )

    cmd = build_base_run_command(
        runtime,
        nofile_limit,
        network_name,
        container_name,
        start_immediately=True,
    )
    maybe_append_amd_gpu_passthrough(cmd, enable_amd_gpu)
    append_host_gateway(cmd, runtime)
    append_mount_specs(cmd, mount_specs)
    if not minimal:
        cmd.extend(["-p", f"{stack_layout.dashboard_port}:8700"])
    append_port_mappings(
        cmd,
        _legacy_runtime_port_mappings(listeners, stack_layout),
    )
    append_env_vars(cmd, legacy_env)
    cmd.append(image)
    return cmd


def _build_router_runtime_command(
    *,
    runtime: str,
    router_image: str,
    nofile_limit: int,
    runtime_network_name: str,
    normalized_platform: str,
    common_env: dict[str, str],
    runtime_paths: dict[str, str],
    setup_mode: bool,
    stack_layout: RuntimeStackLayout,
):
    return _build_service_run_command(
        runtime=runtime,
        image=router_image,
        container_name=stack_layout.router_container_name,
        nofile_limit=nofile_limit,
        network_name=runtime_network_name,
        env_vars=common_env,
        mount_specs=_runtime_mount_specs(runtime_paths, include_models=True),
        port_mappings=[
            (stack_layout.router_port, 50051),
            (stack_layout.metrics_port, 9190),
            (stack_layout.api_port, 8080),
        ],
        entrypoint="/app/start-router.sh",
        command_args=[
            common_env.get("VLLM_SR_RUNTIME_CONFIG_PATH", "/app/config.yaml"),
            "/app/.vllm-sr",
        ],
        enable_amd_gpu=normalized_platform == PLATFORM_AMD,
        start_immediately=not setup_mode,
    )


def _build_envoy_runtime_command(
    *,
    runtime: str,
    envoy_image: str,
    nofile_limit: int,
    runtime_network_name: str,
    common_env: dict[str, str],
    listeners,
    runtime_paths: dict[str, str],
    setup_mode: bool,
    stack_layout: RuntimeStackLayout,
):
    return _build_service_run_command(
        runtime=runtime,
        image=envoy_image,
        container_name=stack_layout.envoy_container_name,
        nofile_limit=nofile_limit,
        network_name=runtime_network_name,
        env_vars={},
        mount_specs=[
            f"{runtime_paths['envoy_config_path']}:/etc/envoy/envoy.yaml:z",
        ],
        port_mappings=[
            (listener["port"] + stack_layout.port_offset, listener["port"])
            for listener in listeners
            if listener.get("port")
        ],
        entrypoint="/usr/local/bin/envoy",
        command_args=[
            "-c",
            "/etc/envoy/envoy.yaml",
            "--log-level",
            "debug",
        ],
        start_immediately=not setup_mode,
    )


def _build_dashboard_runtime_command(
    *,
    runtime: str,
    dashboard_image: str,
    nofile_limit: int,
    runtime_network_name: str,
    common_env: dict[str, str],
    config_dir: str,
    listener_port: int,
    openclaw_network_name: str | None,
    runtime_paths: dict[str, str],
    stack_layout: RuntimeStackLayout,
):
    dashboard_env = _build_dashboard_runtime_env(
        common_env=common_env,
        listener_port=listener_port,
        stack_layout=stack_layout,
    )
    dashboard_mount_specs = _runtime_mount_specs(
        runtime_paths, include_dashboard_data=True
    )
    configure_openclaw_support(
        dashboard_mount_specs,
        dashboard_env,
        config_dir,
        openclaw_network_name,
        runtime,
        stack_layout,
        resolve_docker_cli=resolve_docker_cli_path,
    )
    return _build_service_run_command(
        runtime=runtime,
        image=dashboard_image,
        container_name=stack_layout.dashboard_container_name,
        nofile_limit=nofile_limit,
        network_name=runtime_network_name,
        env_vars=dashboard_env,
        mount_specs=dashboard_mount_specs,
        port_mappings=[(stack_layout.dashboard_port, 8700)],
        entrypoint="/app/start-dashboard.sh",
        command_args=["/app/config.yaml"],
    )


def _build_dashboard_runtime_env(
    *,
    common_env: dict[str, str],
    listener_port: int,
    stack_layout: RuntimeStackLayout,
):
    dashboard_env = dict(common_env)
    dashboard_env.setdefault(
        "TARGET_ROUTER_API_URL", stack_layout.router_api_service_url
    )
    dashboard_env.setdefault(
        "TARGET_ROUTER_METRICS_URL", stack_layout.router_metrics_service_url
    )
    dashboard_env.setdefault(
        "TARGET_ENVOY_URL", stack_layout.envoy_listener_service_url(listener_port)
    )
    dashboard_env.setdefault(
        "TARGET_ENVOY_ADMIN_URL", stack_layout.envoy_admin_service_url
    )
    dashboard_env.setdefault(
        "ENVOY_EXTPROC_ADDRESS", stack_layout.router_container_name
    )
    dashboard_env.setdefault(
        "ENVOY_ROUTER_API_ADDRESS", stack_layout.router_container_name
    )
    dashboard_env.setdefault("VLLM_SR_ENVOY_CONFIG_PATH", "/app/.vllm-sr/envoy.yaml")
    dashboard_env.setdefault(
        "OPENCLAW_DASHBOARD_CONTAINER_NAME", stack_layout.dashboard_container_name
    )
    dashboard_env.setdefault(
        "OPENCLAW_MODEL_GATEWAY_CONTAINER_NAME", stack_layout.envoy_container_name
    )
    return dashboard_env


def _build_legacy_runtime_env(
    *,
    env_vars: dict[str, str],
    minimal: bool,
    mount_specs: list[str],
    config_dir: str,
    openclaw_network_name: str | None,
    runtime: str,
    stack_layout: RuntimeStackLayout,
):
    legacy_env = dict(env_vars)
    if minimal:
        return legacy_env

    legacy_env.setdefault(
        "OPENCLAW_DASHBOARD_CONTAINER_NAME", stack_layout.container_name
    )
    legacy_env.setdefault(
        "OPENCLAW_MODEL_GATEWAY_CONTAINER_NAME", stack_layout.container_name
    )
    configure_openclaw_support(
        mount_specs,
        legacy_env,
        config_dir,
        openclaw_network_name,
        runtime,
        stack_layout,
        resolve_docker_cli=resolve_docker_cli_path,
    )
    return legacy_env


def _legacy_runtime_port_mappings(listeners, stack_layout: RuntimeStackLayout):
    port_mappings = [
        (stack_layout.router_port, 50051),
        (stack_layout.metrics_port, 9190),
        (stack_layout.api_port, 8080),
    ]
    port_mappings.extend(
        (port + stack_layout.port_offset, port)
        for listener in listeners
        if (port := listener.get("port"))
    )
    return port_mappings


def _resolve_platform(env_vars):
    platform = (
        env_vars.get("DASHBOARD_PLATFORM")
        or env_vars.get("VLLM_SR_PLATFORM")
        or os.getenv("VLLM_SR_PLATFORM")
    )
    return _normalize_platform(platform)


def _resolve_nofile_limit():
    nofile_limit = int(os.getenv("VLLM_SR_NOFILE_LIMIT", DEFAULT_NOFILE_LIMIT))
    if nofile_limit < MIN_NOFILE_LIMIT:
        log.warning(
            f"File descriptor limit {nofile_limit} is below minimum {MIN_NOFILE_LIMIT}. "
            "Using minimum value."
        )
        return MIN_NOFILE_LIMIT
    if nofile_limit != DEFAULT_NOFILE_LIMIT:
        log.info(f"Using custom file descriptor limit: {nofile_limit}")
    return nofile_limit


def _prepare_runtime_paths(
    config_file,
    runtime_config_file=None,
    state_root_dir=None,
):
    source_config_path = os.path.abspath(config_file)
    runtime_config_path = os.path.abspath(runtime_config_file or config_file)
    config_dir = (
        os.path.abspath(state_root_dir)
        if state_root_dir
        else os.path.dirname(source_config_path)
    )

    vllm_sr_dir = os.path.join(config_dir, ".vllm-sr")
    os.makedirs(vllm_sr_dir, exist_ok=True)
    log.info(f"Mounting .vllm-sr directory: {vllm_sr_dir}")

    models_dir = os.path.join(config_dir, "models")
    os.makedirs(models_dir, exist_ok=True)

    dashboard_data_dir = os.path.join(config_dir, ".vllm-sr", "dashboard-data")
    os.makedirs(dashboard_data_dir, exist_ok=True)
    log.info(f"Mounting dashboard data directory: {dashboard_data_dir}")

    effective_config_path = runtime_config_path
    envoy_config_path = os.path.join(vllm_sr_dir, "envoy.yaml")

    runtime_container_config = None
    if runtime_config_path != source_config_path:
        runtime_container_config = (
            f"/app/.vllm-sr/{os.path.basename(runtime_config_path)}"
        )
        log.info(
            "Using source config %s with runtime override %s",
            source_config_path,
            runtime_container_config,
        )

    return (
        config_dir,
        {
            "source_config_path": source_config_path,
            "effective_config_path": effective_config_path,
            "vllm_sr_dir": vllm_sr_dir,
            "models_dir": models_dir,
            "dashboard_data_dir": dashboard_data_dir,
            "envoy_config_path": envoy_config_path,
        },
        runtime_container_config,
    )


def _runtime_mount_specs(
    runtime_paths: dict[str, str],
    *,
    include_models: bool = False,
    include_dashboard_data: bool = False,
):
    mounts = [
        f"{runtime_paths['source_config_path']}:/app/config.yaml:z",
        f"{runtime_paths['vllm_sr_dir']}:/app/.vllm-sr:z",
    ]
    if include_models:
        mounts.append(f"{runtime_paths['models_dir']}:/app/models:z")
    if include_dashboard_data:
        mounts.append(f"{runtime_paths['dashboard_data_dir']}:/app/data:z")
    return mounts


def _primary_listener_port(listeners):
    for listener in listeners:
        port = listener.get("port")
        if port:
            return port
    return 8888


def _render_split_envoy_config(
    config_path: str,
    output_path: str,
    stack_layout: RuntimeStackLayout,
) -> None:
    original_extproc = os.environ.get("ENVOY_EXTPROC_ADDRESS")
    original_router_api = os.environ.get("ENVOY_ROUTER_API_ADDRESS")
    os.environ["ENVOY_EXTPROC_ADDRESS"] = stack_layout.router_container_name
    os.environ["ENVOY_ROUTER_API_ADDRESS"] = stack_layout.router_container_name
    try:
        generate_envoy_config_from_user_config(
            parse_user_config(config_path),
            output_path,
        )
        log.info(f"Rendered split Envoy config: {output_path}")
    finally:
        _restore_env_var("ENVOY_EXTPROC_ADDRESS", original_extproc)
        _restore_env_var("ENVOY_ROUTER_API_ADDRESS", original_router_api)


def _restore_env_var(name: str, original_value: str | None) -> None:
    if original_value is None:
        os.environ.pop(name, None)
        return
    os.environ[name] = original_value


def _build_service_run_command(
    *,
    runtime: str,
    image: str,
    container_name: str,
    nofile_limit: int,
    network_name: str,
    env_vars: dict[str, str],
    mount_specs: list[str],
    port_mappings: list[tuple[int, int]],
    entrypoint: str,
    command_args: list[str],
    enable_amd_gpu: bool = False,
    start_immediately: bool = True,
):
    cmd = build_base_run_command(
        runtime,
        nofile_limit,
        network_name,
        container_name,
        start_immediately=start_immediately,
    )
    maybe_append_amd_gpu_passthrough(cmd, enable_amd_gpu)
    append_host_gateway(cmd, runtime)
    append_mount_specs(cmd, mount_specs)
    append_port_mappings(cmd, port_mappings)
    cmd.extend(["--entrypoint", entrypoint])
    append_env_vars(cmd, env_vars)
    cmd.append(image)
    cmd.extend(command_args)
    return cmd


def _cleanup_started_containers(container_names: list[str]) -> None:
    for container_name in reversed(container_names):
        if docker_container_status(container_name) == "running":
            docker_stop_container(container_name)
        docker_remove_container(container_name)



"""Helpers for assembling Docker run/create commands."""

import os

from cli.consts import PLATFORM_AMD
from cli.docker_images import _normalize_platform
from cli.utils import get_logger

log = get_logger(__name__)


def build_base_run_command(
    runtime, nofile_limit, network_name, container_name, *, start_immediately
):
    cmd = [runtime]
    if start_immediately:
        cmd.extend(["run", "-d"])
    else:
        cmd.append("create")
    cmd.extend(
        [
            "--name",
            container_name,
            "--ulimit",
            f"nofile={nofile_limit}:{nofile_limit}",
        ]
    )
    if network_name:
        cmd.extend(["--network", network_name])
    return cmd


def append_amd_gpu_passthrough(cmd, normalized_platform):
    if normalized_platform != PLATFORM_AMD:
        return

    passthrough_enabled = os.getenv("VLLM_SR_AMD_GPU_PASSTHROUGH", "1").lower()
    if passthrough_enabled in ["0", "false", "no", "off"]:
        log.info(
            "AMD GPU passthrough disabled by VLLM_SR_AMD_GPU_PASSTHROUGH environment variable"
        )
        return

    required_devices = ["/dev/kfd", "/dev/dri"]
    mounted_devices = []
    missing_devices = []
    for device in required_devices:
        if os.path.exists(device):
            cmd.extend(["--device", device])
            mounted_devices.append(device)
        else:
            missing_devices.append(device)

    if mounted_devices:
        cmd.extend(["--group-add", "video"])
        cmd.extend(["--cap-add", "SYS_PTRACE"])
        cmd.extend(["--security-opt", "seccomp=unconfined"])
        log.info(
            f"AMD GPU passthrough enabled with devices: {', '.join(mounted_devices)}"
        )

    if missing_devices:
        log.warning(
            "Platform 'amd' selected but missing AMD GPU devices on host: "
            f"{', '.join(missing_devices)}. Container may fall back to CPU."
        )


def append_host_gateway(cmd, runtime):
    if runtime == "docker":
        cmd.append("--add-host=host.docker.internal:host-gateway")


def append_mount_specs(cmd, mount_specs: list[str]):
    for mount_spec in mount_specs:
        cmd.extend(["-v", mount_spec])


def append_port_mappings(cmd, port_mappings: list[tuple[int, int]]):
    for host_port, container_port in port_mappings:
        cmd.extend(["-p", f"{host_port}:{container_port}"])


def append_env_vars(cmd, env_vars: dict[str, str]):
    for key, value in env_vars.items():
        cmd.extend(["-e", f"{key}={value}"])


def maybe_append_amd_gpu_passthrough(cmd, enable_amd_gpu: bool):
    if enable_amd_gpu:
        append_amd_gpu_passthrough(cmd, _normalize_platform(PLATFORM_AMD))



"""Pydantic models for vLLM Semantic Router configuration."""

from enum import Enum
from typing import Any, Dict, List, Literal, Optional

from pydantic import BaseModel, Field, model_validator

from .algorithms import AlgorithmConfig, ModelRef


class Listener(BaseModel):
    """Network listener configuration."""

    name: str
    address: str
    port: int
    timeout: Optional[str] = "300s"


class KeywordSignal(BaseModel):
    """Keyword-based signal configuration."""

    name: str
    operator: str
    keywords: List[str]
    case_sensitive: bool = False


class EmbeddingSignal(BaseModel):
    """Embedding-based signal configuration."""

    name: str
    threshold: float
    candidates: List[str]
    aggregation_method: str = "max"


class ProjectionPartition(BaseModel):
    """Partition metadata coordinating mutually exclusive routing signals."""

    name: str
    semantics: str
    members: List[str]
    temperature: Optional[float] = None
    default: Optional[str] = None


class ProjectionScoreInput(BaseModel):
    """One weighted signal contribution to a derived projection score."""

    type: str
    name: Optional[str] = None
    classifier: Optional[str] = None
    metric: Optional[str] = None
    weight: float
    value_source: Optional[str] = None
    match: Optional[float] = None
    miss: Optional[float] = None


class ProjectionScore(BaseModel):
    """Weighted derived score over existing routing signals."""

    name: str
    method: str
    inputs: List[ProjectionScoreInput]


class ProjectionMappingCalibration(BaseModel):
    """Confidence calibration for a projection mapping output band."""

    method: str
    slope: Optional[float] = None


class ProjectionMappingOutput(BaseModel):
    """One named threshold band emitted by a projection mapping."""

    name: str
    lt: Optional[float] = None
    lte: Optional[float] = None
    gt: Optional[float] = None
    gte: Optional[float] = None


class ProjectionMapping(BaseModel):
    """Maps a derived score into named routing outputs."""

    name: str
    source: str
    method: str
    calibration: Optional[ProjectionMappingCalibration] = None
    outputs: List[ProjectionMappingOutput]


class Projections(BaseModel):
    """Derived routing surfaces that sit alongside base signals."""

    partitions: Optional[List[ProjectionPartition]] = []
    scores: Optional[List[ProjectionScore]] = []
    mappings: Optional[List[ProjectionMapping]] = []


class Domain(BaseModel):
    """Domain category configuration."""

    name: str
    description: str
    mmlu_categories: Optional[List[str]] = None


class FactCheck(BaseModel):
    """Fact-checking signal configuration."""

    name: str
    description: str


class UserFeedback(BaseModel):
    """User feedback signal configuration."""

    name: str
    description: str


class Preference(BaseModel):
    """Route preference signal configuration."""

    name: str
    description: str
    threshold: Optional[float] = None
    examples: Optional[List[str]] = None


class Language(BaseModel):
    """Language detection signal configuration."""

    name: str
    description: str


class ContextRule(BaseModel):
    """Context-based (token count) signal configuration."""

    name: str
    min_tokens: str  # Supports suffixes: "1K", "1.5M", etc.
    max_tokens: str
    description: Optional[str] = None


class StructureSource(BaseModel):
    """Source selector for structure-based signals."""

    type: str
    pattern: Optional[str] = None
    keywords: Optional[List[str]] = None
    case_sensitive: bool = False
    sequences: Optional[List[List[str]]] = None

    class Config:
        extra = "forbid"


class StructureFeature(BaseModel):
    """Typed request-shape feature extractor."""

    type: str
    source: StructureSource

    class Config:
        extra = "forbid"


class NumericPredicate(BaseModel):
    """Numeric threshold predicate for structure signals."""

    gt: Optional[float] = None
    gte: Optional[float] = None
    lt: Optional[float] = None
    lte: Optional[float] = None

    class Config:
        extra = "forbid"


class StructureRule(BaseModel):
    """Request-shape routing signal configuration."""

    name: str
    description: Optional[str] = None
    feature: StructureFeature
    predicate: Optional[NumericPredicate] = None

    class Config:
        extra = "forbid"


class ComplexityCandidates(BaseModel):
    """Complexity candidates configuration."""

    candidates: List[str]
    image_candidates: Optional[List[str]] = None


class PrototypeScoringConfig(BaseModel):
    """Prototype-bank construction and scoring controls for embedding-backed signals."""

    enabled: Optional[bool] = None
    cluster_similarity_threshold: Optional[float] = None
    max_prototypes: Optional[int] = None
    best_weight: Optional[float] = None
    top_m: Optional[int] = None
    margin_threshold: Optional[float] = None


class EmbeddingClassifierConfig(BaseModel):
    """Embedding classifier tuning, including prototype-aware label scoring controls."""

    model_type: Optional[str] = None
    preload_embeddings: Optional[bool] = None
    target_dimension: Optional[int] = None
    target_layer: Optional[int] = None
    enable_soft_matching: Optional[bool] = None
    top_k: Optional[int] = None
    min_score_threshold: Optional[float] = None
    prototype_scoring: Optional[PrototypeScoringConfig] = None

    class Config:
        extra = "allow"


class ComplexityRule(BaseModel):
    """Complexity-based signal configuration using embedding similarity.

    The composer field allows filtering based on other signals (e.g., only apply
    code_complexity when domain is "computer science"). This is evaluated after
    all signals are computed in parallel, enabling signal dependencies.
    """

    name: str
    threshold: float = 0.1
    hard: ComplexityCandidates
    easy: ComplexityCandidates
    description: Optional[str] = None
    composer: Optional["Rules"] = None  # Forward reference, defined below


class JailbreakRule(BaseModel):
    """Jailbreak detection signal configuration.

    Supports two methods:
    - "classifier" (default): BERT/LoRA-based jailbreak classifier
    - "contrastive": Embedding-based contrastive scoring against jailbreak/benign KBs
    """

    name: str
    threshold: float
    method: Optional[str] = None  # "classifier" (default) or "contrastive"
    include_history: bool = False
    jailbreak_patterns: Optional[list[str]] = (
        None  # Known jailbreak prompts (contrastive KB)
    )
    benign_patterns: Optional[list[str]] = None  # Known benign prompts (contrastive KB)
    description: Optional[str] = None


class PIIRule(BaseModel):
    """PII detection signal configuration."""

    name: str
    threshold: float
    pii_types_allowed: Optional[List[str]] = None
    include_history: bool = False
    description: Optional[str] = None


class ModalityRule(BaseModel):
    """Modality detection signal configuration.

    Classifies whether a prompt requires text (AR), image (DIFFUSION), or both (BOTH).
    Detection configuration is read from modality_detector (InlineModels).
    """

    name: str
    description: Optional[str] = None


class Subject(BaseModel):
    """RBAC subject (user or group) for role binding."""

    kind: str  # "User" or "Group"
    name: str


class RoleBindingRule(BaseModel):
    """RBAC role binding signal configuration.

    Maps subjects (users/groups) to a named role following the Kubernetes RBAC pattern.
    The role name is emitted as a signal of type "authz" in the decision engine.
    User identity is read from x-authz-user-id and x-authz-user-groups headers.
    """

    name: str
    role: str
    subjects: List[Subject]
    description: Optional[str] = None


class KBSignalTarget(BaseModel):
    """Binding target for a named knowledge base."""

    kind: Literal["label", "group"]
    value: str


class KBSignal(BaseModel):
    """Knowledge-base signal bound to a named global KB instance."""

    name: str
    kb: str
    target: KBSignalTarget
    match: Optional[Literal["best", "threshold"]] = None


class Reask(BaseModel):
    """History-aware repeated-question dissatisfaction signal."""

    name: str
    description: Optional[str] = None
    threshold: Optional[float] = None
    lookback_turns: Optional[int] = None


class Signals(BaseModel):
    """All signal configurations."""

    keywords: Optional[List[KeywordSignal]] = []
    embeddings: Optional[List[EmbeddingSignal]] = []
    domains: Optional[List[Domain]] = []
    fact_check: Optional[List[FactCheck]] = []
    user_feedbacks: Optional[List[UserFeedback]] = []
    reasks: Optional[List[Reask]] = []
    preferences: Optional[List[Preference]] = []
    language: Optional[List[Language]] = []
    context: Optional[List[ContextRule]] = []
    structure: Optional[List[StructureRule]] = []
    complexity: Optional[List[ComplexityRule]] = []
    modality: Optional[List[ModalityRule]] = []
    role_bindings: Optional[List[RoleBindingRule]] = []
    jailbreak: Optional[List[JailbreakRule]] = []
    pii: Optional[List[PIIRule]] = []
    kb: Optional[List[KBSignal]] = []


class Condition(BaseModel):
    """Routing condition node (leaf or composite boolean expression)."""

    type: Optional[str] = None
    name: Optional[str] = None
    operator: Optional[str] = None
    conditions: Optional[List["Condition"]] = None

    @model_validator(mode="after")
    def validate_node_shape(self):
        has_leaf_fields = self.type is not None or self.name is not None
        has_operator = self.operator is not None

        if has_leaf_fields and has_operator:
            raise ValueError(
                "condition node must be either leaf (type/name) or composite (operator/conditions), not both"
            )

        if has_operator:
            if not self.conditions:
                raise ValueError(
                    "composite condition node requires non-empty conditions"
                )
            op = self.operator.strip().upper()
            if op not in {"AND", "OR", "NOT"}:
                raise ValueError("operator must be one of: AND, OR, NOT")
            if op == "NOT" and len(self.conditions) != 1:
                raise ValueError("NOT operator must have exactly one child condition")
            return self

        # Leaf node validation
        if self.type is None or self.name is None:
            raise ValueError("leaf condition node requires both type and name")
        if self.conditions:
            raise ValueError("leaf condition node cannot define child conditions")
        return self


class Rules(BaseModel):
    """Routing rules.

    Accepts three formats:
    1. Composite: {operator: "AND", conditions: [...]}
    2. Match-all: {operator: "AND"} or {} (no WHEN clause)
    3. Leaf node: {type: "keyword", name: "x"} (single signal ref)

    Formats 2 and 3 are auto-normalised to composite form.
    """

    operator: str = "AND"
    conditions: List[Condition] = []

    @model_validator(mode="before")
    @classmethod
    def normalise_leaf_or_empty(cls, data):
        """Wrap a bare leaf node into AND([leaf]) and fill missing fields."""
        if not isinstance(data, dict):
            return data
        # Leaf node: has type/name but no operator → wrap in AND
        if "type" in data and "operator" not in data:
            leaf = {"type": data["type"], "name": data.get("name", "")}
            return {"operator": "AND", "conditions": [leaf]}
        return data


class PluginType(str, Enum):
    """Supported plugin types."""

    SEMANTIC_CACHE = "semantic-cache"
    SYSTEM_PROMPT = "system_prompt"
    HEADER_MUTATION = "header_mutation"
    HALLUCINATION = "hallucination"
    ROUTER_REPLAY = "router_replay"
    MEMORY = "memory"
    RAG = "rag"
    IMAGE_GEN = "image_gen"
    FAST_RESPONSE = "fast_response"
    REQUEST_PARAMS = "request_params"
    RESPONSE_JAILBREAK = "response_jailbreak"
    TOOLS = "tools"


class SemanticCachePluginConfig(BaseModel):
    """Configuration for semantic-cache plugin."""

    enabled: bool
    similarity_threshold: Optional[float] = Field(
        default=None,
        ge=0.0,
        le=1.0,
        description="Similarity threshold (0.0-1.0, default: None)",
    )
    ttl_seconds: Optional[int] = Field(
        default=None, ge=0, description="TTL in seconds (must be >= 0, default: None)"
    )


class FastResponsePluginConfig(BaseModel):
    """Configuration for fast_response plugin."""

    message: str


class RequestParamsPluginConfig(BaseModel):
    """Configuration for request_params plugin."""

    blocked_params: Optional[List[str]] = None
    max_tokens_limit: Optional[int] = Field(default=None, ge=1)
    max_n: Optional[int] = Field(default=None, ge=1)
    strip_unknown: Optional[bool] = None


class ResponseJailbreakPluginConfig(BaseModel):
    """Configuration for response_jailbreak plugin."""

    enabled: bool
    threshold: Optional[float] = Field(default=None, ge=0.0, le=1.0)
    action: Optional[Literal["block", "header", "none"]] = None


class ToolsPluginConfig(BaseModel):
    """Configuration for tools plugin."""

    enabled: bool
    mode: Literal["none", "passthrough", "filtered"] = "passthrough"
    semantic_selection: Optional[bool] = None
    allow_tools: Optional[List[str]] = None
    block_tools: Optional[List[str]] = None


class SystemPromptPluginConfig(BaseModel):
    """Configuration for system_prompt plugin."""

    enabled: Optional[bool] = None
    system_prompt: Optional[str] = None
    mode: Optional[Literal["replace", "insert"]] = None


class HeaderPair(BaseModel):
    """Header name-value pair."""

    name: str
    value: str


class HeaderMutationPluginConfig(BaseModel):
    """Configuration for header_mutation plugin."""

    add: Optional[List[HeaderPair]] = None
    update: Optional[List[HeaderPair]] = None
    delete: Optional[List[str]] = None


class HallucinationPluginConfig(BaseModel):
    """Configuration for hallucination plugin."""

    enabled: bool
    use_nli: Optional[bool] = None
    hallucination_action: Optional[Literal["header", "body", "none"]] = None
    unverified_factual_action: Optional[Literal["header", "body", "none"]] = None
    include_hallucination_details: Optional[bool] = None


class RouterReplayPluginConfig(BaseModel):
    """Configuration for router_replay plugin.

    The router_replay plugin captures routing decisions and payload snippets
    for later debugging and replay. Records are stored in memory and accessible
    via the /v1/router_replay API endpoint.
    """

    enabled: bool = True
    max_records: int = Field(
        default=10000,
        gt=0,
        description="Maximum records in memory (must be > 0, default: 10000)",
    )
    capture_request_body: bool = True  # Capture request payloads
    capture_response_body: bool = True  # Capture response payloads
    max_body_bytes: int = Field(
        default=4096,
        gt=0,
        description="Max bytes to capture per body (must be > 0, default: 4096)",
    )


class MemoryPluginConfig(BaseModel):
    """Configuration for memory plugin (per-decision memory settings)."""

    enabled: bool = True
    retrieval_limit: Optional[int] = Field(
        default=None,
        gt=0,
        description="Max memories to retrieve (default: use global config)",
    )
    similarity_threshold: Optional[float] = Field(
        default=None,
        ge=0.0,
        le=1.0,
        description="Min similarity score (0.0-1.0, default: use global config)",
    )
    auto_store: Optional[bool] = Field(
        default=None,
        description="Auto-extract memories from conversation (default: use request config)",
    )


class RAGPluginConfig(BaseModel):
    """Configuration for RAG (Retrieval-Augmented Generation) plugin.

    The RAG plugin retrieves relevant context from external knowledge bases
    and injects it into the LLM request.

    Supported backends:
    - milvus: Milvus vector database (reuses semantic cache connection)
    - external_api: External REST API (OpenAI, Pinecone, Weaviate, Elasticsearch)
    - mcp: MCP tool-based retrieval
    - openai: OpenAI file_search with vector stores
    - hybrid: Multi-backend with fallback strategy
    """

    # Required: Enable RAG retrieval
    enabled: bool = Field(..., description="Enable RAG retrieval for this decision")

    # Required: Backend type (milvus, external_api, mcp, openai, hybrid)
    backend: str = Field(
        ...,
        description="Retrieval backend: milvus, external_api, mcp, openai, hybrid",
    )

    # Optional: Similarity threshold (0.0-1.0)
    # Only documents with similarity >= threshold will be retrieved
    similarity_threshold: Optional[float] = Field(
        default=None,
        ge=0.0,
        le=1.0,
        description="Similarity threshold for retrieval (0.0-1.0)",
    )

    # Optional: Number of top-k documents to retrieve
    top_k: Optional[int] = Field(
        default=None,
        ge=1,
        description="Number of top-k documents to retrieve",
    )

    # Optional: Maximum context length (in characters)
    max_context_length: Optional[int] = Field(
        default=None,
        ge=1,
        description="Maximum context length to inject (characters)",
    )

    # Optional: Context injection mode
    # - "tool_role": Inject as tool role messages (compatible with hallucination detection)
    # - "system_prompt": Prepend to system prompt
    injection_mode: Optional[Literal["tool_role", "system_prompt"]] = Field(
        default=None,
        description="Injection mode: tool_role (default) or system_prompt",
    )

    # Optional: Backend-specific configuration
    # Structure depends on backend type (see Go: rag_plugin.go lines 64-174)
    backend_config: Optional[Dict[str, Any]] = Field(
        default=None,
        description="Backend-specific configuration",
    )

    # Optional: Fallback behavior on retrieval failure
    # - "skip": Continue without context (default)
    # - "block": Return error response
    # - "warn": Continue with warning header
    on_failure: Optional[Literal["skip", "block", "warn"]] = Field(
        default=None,
        description="On failure: skip (default), block, or warn",
    )

    # Optional: Cache retrieved results
    cache_results: Optional[bool] = Field(
        default=None,
        description="Cache retrieved results",
    )

    # Optional: Cache TTL (seconds)
    cache_ttl_seconds: Optional[int] = Field(
        default=None,
        ge=1,
        description="Cache TTL in seconds",
    )

    # Optional: Minimum confidence for triggering retrieval
    min_confidence_threshold: Optional[float] = Field(
        default=None,
        ge=0.0,
        le=1.0,
        description="Minimum confidence for triggering retrieval",
    )


class PluginConfig(BaseModel):
    """Plugin configuration with type validation.

    Configuration schema validation is performed in the validator module
    to ensure proper plugin-specific validation.
    """

    type: PluginType
    configuration: Dict[str, Any]

    def model_dump(self, **kwargs):
        """Override model_dump to serialize PluginType enum as string value."""
        # Use mode='python' to get Python native types, then convert enum
        # Pop mode from kwargs to avoid duplicate argument if caller passes it
        mode = kwargs.pop("mode", "python")
        data = super().model_dump(mode=mode, **kwargs)
        # Convert PluginType enum to its string value for YAML serialization
        if isinstance(data.get("type"), PluginType):
            data["type"] = data["type"].value
        elif hasattr(data.get("type"), "value"):
            data["type"] = data["type"].value
        return data


class ImageGenPluginConfig(BaseModel):
    """Configuration for image_gen plugin."""

    enabled: bool
    backend: str
    backend_config: Optional[Dict[str, Any]] = None
    modality_detection: Optional[Dict[str, Any]] = None
    default_width: Optional[int] = Field(default=None, ge=1)
    default_height: Optional[int] = Field(default=None, ge=1)
    max_inference_steps: Optional[int] = Field(default=None, ge=1)
    timeout_seconds: Optional[int] = Field(default=None, ge=1)


class Decision(BaseModel):
    """Routing decision configuration."""

    name: str
    description: str
    priority: int
    rules: Rules
    modelRefs: List[ModelRef] = Field(alias="modelRefs")
    algorithm: Optional[AlgorithmConfig] = None  # Multi-model orchestration algorithm
    plugins: Optional[List[PluginConfig]] = []

    class Config:
        populate_by_name = True


class ModelPricing(BaseModel):
    """Model pricing configuration."""

    currency: Optional[str] = "USD"
    prompt_per_1m: Optional[float] = 0.0
    completion_per_1m: Optional[float] = 0.0


class Model(BaseModel):
    """Provider model binding for canonical providers.models entries."""

    name: str
    reasoning_family: Optional[str] = None
    provider_model_id: Optional[str] = None
    backend_refs: List["BackendRef"] = Field(default_factory=list)
    pricing: Optional[ModelPricing] = None
    api_format: Optional[str] = None
    external_model_ids: Optional[Dict[str, str]] = None


class LoRAAdapter(BaseModel):
    """LoRA adapter metadata exposed under routing.modelCards[].loras."""

    name: str
    description: Optional[str] = None


class RoutingModel(BaseModel):
    """Semantic model catalog entry exposed to routing/DSL."""

    name: str
    param_size: Optional[str] = None
    context_window_size: Optional[int] = Field(default=None, ge=1)
    description: Optional[str] = None
    capabilities: Optional[List[str]] = None
    loras: Optional[List[LoRAAdapter]] = None
    tags: Optional[List[str]] = None
    quality_score: Optional[float] = Field(default=None, ge=0, le=1)
    modality: Optional[str] = None


class ReasoningFamily(BaseModel):
    """Reasoning family configuration."""

    type: str
    parameter: str


class BackendRef(BaseModel):
    """Inline backend access details carried under providers.models[].backend_refs."""

    name: Optional[str] = None
    endpoint: Optional[str] = None
    protocol: str = "http"
    weight: int = 1
    type: Optional[str] = None
    base_url: Optional[str] = None
    provider: Optional[str] = None
    auth_header: Optional[str] = None
    auth_prefix: Optional[str] = None
    extra_headers: Optional[Dict[str, str]] = None
    api_version: Optional[str] = None
    chat_path: Optional[str] = None
    api_key: Optional[str] = None
    api_key_env: Optional[str] = None

    def resolve_api_key(self) -> Optional[str]:
        if self.api_key:
            return self.api_key
        if self.api_key_env:
            import os

            return os.getenv(self.api_key_env)
        return None


class ProviderDefaults(BaseModel):
    """Provider-wide defaults that should not be mixed into per-model access bindings."""

    default_model: Optional[str] = None
    reasoning_families: Optional[Dict[str, "ReasoningFamily"]] = Field(
        default_factory=dict
    )
    default_reasoning_effort: Optional[str] = "high"


class Providers(BaseModel):
    """Provider configuration."""

    defaults: ProviderDefaults = Field(default_factory=ProviderDefaults)
    models: List[Model] = Field(default_factory=list)

    @property
    def default_model(self) -> Optional[str]:
        return self.defaults.default_model

    @property
    def reasoning_families(self) -> Dict[str, "ReasoningFamily"]:
        return self.defaults.reasoning_families or {}

    @property
    def default_reasoning_effort(self) -> Optional[str]:
        return self.defaults.default_reasoning_effort


class Routing(BaseModel):
    """Canonical routing block."""

    model_cards: List[RoutingModel] = Field(default_factory=list, alias="modelCards")
    signals: Signals = Field(default_factory=Signals)
    projections: Projections = Field(default_factory=Projections)
    decisions: List[Decision] = Field(default_factory=list)

    class Config:
        populate_by_name = True


class EmbeddingModelsConfig(BaseModel):
    """Embedding models configuration for memory and semantic features."""

    qwen3_model_path: Optional[str] = Field(
        None, description="Path to Qwen3-Embedding model"
    )
    gemma_model_path: Optional[str] = Field(
        None, description="Path to EmbeddingGemma model"
    )
    mmbert_model_path: Optional[str] = Field(
        None, description="Path to mmBERT 2D Matryoshka model"
    )
    multimodal_model_path: Optional[str] = Field(
        None,
        description="Path to multi-modal embedding model (text/image/audio)",
    )
    bert_model_path: Optional[str] = Field(
        None,
        description="Path to BERT/MiniLM model (recommended for memory retrieval)",
    )
    embedding_config: Optional[EmbeddingClassifierConfig] = Field(
        default=None,
        description="Embedding classifier tuning (model_type/target_dimension/top_k/prototype_scoring/etc.)",
    )
    use_cpu: bool = Field(True, description="Use CPU for inference")

    class Config:
        # Preserve advanced nested fields when users pass through custom config blocks.
        extra = "allow"


class UserConfig(BaseModel):
    """Canonical v0.3 user configuration."""

    version: str
    listeners: List[Listener] = Field(default_factory=list)
    providers: Providers = Field(default_factory=Providers)
    routing: Routing = Field(default_factory=Routing)
    global_: Optional[Dict[str, Any]] = Field(default=None, alias="global")
    setup: Optional[Dict[str, Any]] = None

    @property
    def signals(self) -> Signals:
        return self.routing.signals

    @property
    def decisions(self) -> List[Decision]:
        return self.routing.decisions

    class Config:
        populate_by_name = True
        extra = "forbid"


# Resolve forward references for recursive condition trees.
Condition.model_rebuild()
Model.model_rebuild()



"""Shared CLI config-contract inventories and helpers."""

from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass
from typing import Any

CANONICAL_VERSION = "v0.3"

CANONICAL_TOP_LEVEL_KEYS = frozenset(
    {
        "version",
        "listeners",
        "providers",
        "routing",
        "global",
        "setup",
    }
)

LEGACY_PROVIDER_DEFAULT_KEYS = (
    "default_model",
    "reasoning_families",
    "default_reasoning_effort",
)

LEGACY_PROVIDER_KEYS = frozenset(
    {
        *LEGACY_PROVIDER_DEFAULT_KEYS,
        "model_config",
        "vllm_endpoints",
        "provider_profiles",
    }
)

LEGACY_PROVIDER_MODEL_SURFACE_KEYS = frozenset(
    {
        "endpoints",
        "access_key",
        "param_size",
        "context_window_size",
        "description",
        "capabilities",
        "loras",
        "quality_score",
        "modality",
        "tags",
    }
)


@dataclass(frozen=True)
class SignalFamilySpec:
    """Canonical inventory for one routing signal family."""

    canonical_key: str
    signal_attr: str
    condition_type: str
    legacy_key: str | None = None
    reference_suffixes: tuple[str, ...] = ()


SIGNAL_FAMILY_SPECS = (
    SignalFamilySpec("keywords", "keywords", "keyword", "keyword_rules"),
    SignalFamilySpec("embeddings", "embeddings", "embedding", "embedding_rules"),
    SignalFamilySpec("domains", "domains", "domain", "categories"),
    SignalFamilySpec("fact_check", "fact_check", "fact_check", "fact_check_rules"),
    SignalFamilySpec(
        "user_feedbacks",
        "user_feedbacks",
        "user_feedback",
        "user_feedback_rules",
    ),
    SignalFamilySpec("reasks", "reasks", "reask", "reask_rules"),
    SignalFamilySpec("preferences", "preferences", "preference", "preference_rules"),
    SignalFamilySpec("language", "language", "language", "language_rules"),
    SignalFamilySpec("context", "context", "context", "context_rules"),
    SignalFamilySpec("structure", "structure", "structure", "structure_rules"),
    SignalFamilySpec(
        "complexity",
        "complexity",
        "complexity",
        "complexity_rules",
        ("easy", "medium", "hard"),
    ),
    SignalFamilySpec("modality", "modality", "modality", "modality_rules"),
    SignalFamilySpec("role_bindings", "role_bindings", "authz", "role_bindings"),
    SignalFamilySpec("jailbreak", "jailbreak", "jailbreak", "jailbreak"),
    SignalFamilySpec("pii", "pii", "pii", "pii"),
    SignalFamilySpec("kb", "kb", "kb", "kb"),
)

LEGACY_SIGNAL_KEY_TO_CANONICAL = {
    spec.legacy_key: spec.canonical_key
    for spec in SIGNAL_FAMILY_SPECS
    if spec.legacy_key is not None
}

LEGACY_ROUTING_KEYS = frozenset(
    {"signals", "decisions", *LEGACY_SIGNAL_KEY_TO_CANONICAL}
)

_SIGNAL_FAMILY_BY_CONDITION_TYPE = {
    spec.condition_type: spec for spec in SIGNAL_FAMILY_SPECS
}


def iter_named_signal_entries(signals: Any) -> Iterable[tuple[str, str]]:
    """Yield canonical signal family keys and declared signal names."""
    if not signals:
        return
    for spec in SIGNAL_FAMILY_SPECS:
        for signal in getattr(signals, spec.signal_attr, None) or []:
            name = getattr(signal, "name", None)
            if name:
                yield spec.canonical_key, name


def build_signal_reference_index(signals: Any) -> set[str]:
    """Build the valid decision reference names for declared signals."""
    names: set[str] = set()
    if not signals:
        return names

    for spec in SIGNAL_FAMILY_SPECS:
        for signal in getattr(signals, spec.signal_attr, None) or []:
            name = getattr(signal, "name", None)
            if not name:
                continue
            if spec.reference_suffixes:
                for suffix in spec.reference_suffixes:
                    names.add(f"{name}:{suffix}")
                continue
            names.add(name)

    return names


def build_projection_reference_index(projections: Any) -> set[str]:
    """Build the valid decision reference names for declared projection outputs."""
    names: set[str] = set()
    if not projections:
        return names

    for mapping in getattr(projections, "mappings", None) or []:
        for output in getattr(mapping, "outputs", None) or []:
            name = getattr(output, "name", None)
            if name:
                names.add(name)

    return names


def is_signal_condition_type(condition_type: str | None) -> bool:
    """Return whether a decision condition type references a routing signal."""
    if not condition_type:
        return False
    return condition_type.strip().lower() in _SIGNAL_FAMILY_BY_CONDITION_TYPE


def signal_reference_exists(
    signal_names: set[str], condition_type: str | None, raw_name: str | None
) -> bool:
    """Return whether a decision condition references a known signal."""
    if not raw_name or not is_signal_condition_type(condition_type):
        return False

    normalized_type = condition_type.strip().lower()
    spec = _SIGNAL_FAMILY_BY_CONDITION_TYPE[normalized_type]
    if spec.reference_suffixes:
        return raw_name in signal_names
    return raw_name.split(":", 1)[0] in signal_names



"""CLI-local canonical defaults for router service storage backends."""

from __future__ import annotations

from collections.abc import Mapping
from typing import Any

from cli.runtime_stack import RuntimeStackLayout
from cli.utils import get_logger

log = get_logger(__name__)

CANONICAL_SERVICE_DEFAULTS: dict[str, dict[str, object]] = {
    "response_api": {
        "enabled": True,
        "store_backend": "redis",
    },
    "router_replay": {
        "enabled": True,
        "store_backend": "postgres",
    },
}

_INVALID_MAPPING = object()


def is_setup_mode_config(config: Mapping[str, Any]) -> bool:
    """Return True when the config is a setup-mode bootstrap config."""
    setup_config = config.get("setup")
    return isinstance(setup_config, Mapping) and setup_config.get("mode") is True


def detect_canonical_storage_backends(config: Mapping[str, Any]) -> set[str]:
    """Return provisionable backends implied by canonical service defaults.

    Setup-mode bootstrap still uses the local default sidecar stack so the
    first `vllm-sr serve` starts the same local storage dependencies that the
    activated config will later consume.
    """
    backends: set[str] = set()
    for service_key in CANONICAL_SERVICE_DEFAULTS:
        backend = effective_service_backend(config, service_key)
        if backend in {"redis", "postgres"}:
            backends.add(backend)
    return backends


def effective_service_backend(
    config: Mapping[str, Any], service_key: str
) -> str | None:
    """Return the effective store backend for a router service."""
    service_config = _merged_service_config(config, service_key)
    if service_config is None:
        return None
    if service_config.get("enabled") is False:
        return None

    backend = str(service_config.get("store_backend") or "").strip().lower()
    return backend or None


def inject_local_service_runtime_defaults(
    config: dict[str, object], stack_layout: RuntimeStackLayout
) -> bool:
    """Inject local Docker connection defaults for canonical service backends."""
    if is_setup_mode_config(config):
        return False

    services = _ensure_runtime_services_mapping(config)
    if services is None:
        return False

    changed = False
    for service_key, default_config in CANONICAL_SERVICE_DEFAULTS.items():
        changed = (
            _inject_service_runtime_defaults(
                services,
                service_key,
                default_config,
                stack_layout,
            )
            or changed
        )

    return changed


def _services_mapping(config: Mapping[str, Any]) -> Mapping[str, Any] | object:
    global_config = config.get("global")
    if global_config is None:
        return {}
    if not isinstance(global_config, Mapping):
        log.warning(
            "Skipping canonical service defaults because global is not a mapping"
        )
        return _INVALID_MAPPING

    services = global_config.get("services")
    if services is None:
        return {}
    if not isinstance(services, Mapping):
        log.warning(
            "Skipping canonical service defaults because global.services is not a mapping"
        )
        return _INVALID_MAPPING
    return services


def _merged_service_config(
    config: Mapping[str, Any], service_key: str
) -> dict[str, object] | None:
    defaults = CANONICAL_SERVICE_DEFAULTS.get(service_key)
    if defaults is None:
        return None

    services = _services_mapping(config)
    if services is _INVALID_MAPPING:
        return None

    raw_service = services.get(service_key)
    if raw_service is None:
        return dict(defaults)
    if not isinstance(raw_service, Mapping):
        log.warning(
            "Skipping canonical service defaults for global.services.%s because it is not a mapping",
            service_key,
        )
        return None

    merged = dict(defaults)
    merged.update(dict(raw_service))
    return merged


def _ensure_mapping(
    parent: dict[str, object], key: str, path: str
) -> dict[str, object] | None:
    child = parent.get(key)
    if child is None:
        mapping: dict[str, object] = {}
        parent[key] = mapping
        return mapping
    if not isinstance(child, dict):
        log.warning(
            "Skipping local service default injection because %s is not a mapping",
            path,
        )
        return None
    return child


def _ensure_runtime_services_mapping(
    config: dict[str, object],
) -> dict[str, object] | None:
    global_config = _ensure_mapping(config, "global", "global")
    if global_config is None:
        return None
    return _ensure_mapping(global_config, "services", "global.services")


def _inject_service_runtime_defaults(
    services: dict[str, object],
    service_key: str,
    default_config: dict[str, object],
    stack_layout: RuntimeStackLayout,
) -> bool:
    service_config, changed = _ensure_service_runtime_mapping(services, service_key)
    if service_config is None:
        return changed

    changed = _apply_missing_defaults(service_config, default_config) or changed
    if service_config.get("enabled") is False:
        return changed

    return (
        _inject_backend_runtime_defaults(service_key, service_config, stack_layout)
        or changed
    )


def _ensure_service_runtime_mapping(
    services: dict[str, object], service_key: str
) -> tuple[dict[str, object] | None, bool]:
    service_config = services.get(service_key)
    if service_config is None:
        service_mapping: dict[str, object] = {}
        services[service_key] = service_mapping
        return service_mapping, True
    if not isinstance(service_config, dict):
        log.warning(
            "Skipping local service default injection for global.services.%s because it is not a mapping",
            service_key,
        )
        return None, False
    return service_config, False


def _inject_backend_runtime_defaults(
    service_key: str,
    service_config: dict[str, object],
    stack_layout: RuntimeStackLayout,
) -> bool:
    backend = _normalized_backend_value(service_config.get("store_backend"))
    if not backend:
        return False

    backend_defaults = _local_backend_defaults(
        service_key=service_key,
        backend=backend,
        stack_layout=stack_layout,
    )
    if not backend_defaults:
        return False

    backend_config, changed = _ensure_backend_runtime_mapping(
        service_config,
        service_key,
        backend,
        backend_defaults,
    )
    if backend_config is None:
        return changed

    return _apply_missing_or_blank_defaults(backend_config, backend_defaults) or changed


def _ensure_backend_runtime_mapping(
    service_config: dict[str, object],
    service_key: str,
    backend: str,
    backend_defaults: dict[str, object],
) -> tuple[dict[str, object] | None, bool]:
    backend_config = service_config.get(backend)
    if backend_config is None:
        service_config[backend] = dict(backend_defaults)
        created_config = service_config.get(backend)
        if isinstance(created_config, dict):
            return created_config, True
        return None, True
    if not isinstance(backend_config, dict):
        log.warning(
            "Skipping local service default injection for global.services.%s.%s because it is not a mapping",
            service_key,
            backend,
        )
        return None, False
    return backend_config, False


def _apply_missing_defaults(
    target: dict[str, object], defaults: Mapping[str, object]
) -> bool:
    changed = False
    for key, value in defaults.items():
        if key not in target:
            target[key] = value
            changed = True
    return changed


def _apply_missing_or_blank_defaults(
    target: dict[str, object], defaults: Mapping[str, object]
) -> bool:
    changed = False
    for key, value in defaults.items():
        if key not in target or target[key] in (None, ""):
            target[key] = value
            changed = True
    return changed


def _normalized_backend_value(raw_backend: object) -> str | None:
    backend = str(raw_backend or "").strip().lower()
    return backend or None


def _local_backend_defaults(
    *, service_key: str, backend: str, stack_layout: RuntimeStackLayout
) -> dict[str, object]:
    if backend == "redis":
        return {
            "address": f"{stack_layout.redis_container_name}:6379",
            "db": 0,
        }

    if service_key == "router_replay" and backend == "postgres":
        return {
            "host": stack_layout.postgres_container_name,
            "port": 5432,
            "database": "vsr",
            "user": "router",
            "password": "router-secret",
            "ssl_mode": "disable",
        }

    return {}



import sys
from pathlib import Path

import pytest

PROJECT_ROOT = Path(__file__).resolve().parents[1]
if str(PROJECT_ROOT) not in sys.path:
    sys.path.insert(0, str(PROJECT_ROOT))

from cli import docker_images  # noqa: E402
from cli.consts import (  # noqa: E402
    VLLM_SR_DASHBOARD_DOCKER_IMAGE_DEFAULT,
    VLLM_SR_DOCKER_IMAGE_DEFAULT,
    VLLM_SR_DOCKER_IMAGE_ROCM,
    VLLM_SR_ENVOY_DOCKER_IMAGE_DEFAULT,
)


@pytest.fixture(autouse=True)
def clear_runtime_image_env(monkeypatch):
    for env_name in (
        "VLLM_SR_IMAGE",
        "VLLM_SR_ROUTER_IMAGE",
        "VLLM_SR_ENVOY_IMAGE",
        "VLLM_SR_DASHBOARD_IMAGE",
        "VLLM_SR_PLATFORM",
    ):
        monkeypatch.delenv(env_name, raising=False)


def test_get_runtime_images_falls_back_to_base_image(monkeypatch):
    ensured = []
    monkeypatch.setattr(
        docker_images,
        "_resolve_selected_image",
        lambda image, normalized_platform: "base:latest",
    )
    monkeypatch.setattr(
        docker_images,
        "_ensure_image_available",
        lambda image, pull_policy: ensured.append((image, pull_policy)),
    )

    images = docker_images.get_runtime_images(pull_policy="never")

    assert images == {
        "router": "base:latest",
        "envoy": "base:latest",
        "dashboard": "base:latest",
    }
    assert ensured == [("base:latest", "never")]


def test_get_runtime_images_prefers_service_specific_env_overrides(monkeypatch):
    ensured = []
    monkeypatch.setattr(
        docker_images,
        "_resolve_selected_image",
        lambda image, normalized_platform: "base:latest",
    )
    monkeypatch.setattr(
        docker_images,
        "_ensure_image_available",
        lambda image, pull_policy: ensured.append((image, pull_policy)),
    )
    monkeypatch.setenv("VLLM_SR_ENVOY_IMAGE", "envoy:separate")

    images = docker_images.get_runtime_images(pull_policy="ifnotpresent")

    assert images == {
        "router": "base:latest",
        "envoy": "envoy:separate",
        "dashboard": "base:latest",
    }
    assert ensured == [
        ("base:latest", "ifnotpresent"),
        ("envoy:separate", "ifnotpresent"),
    ]


def test_get_runtime_images_derives_official_envoy_image_from_official_base(
    monkeypatch,
):
    ensured = []
    monkeypatch.setattr(
        docker_images,
        "_resolve_selected_image",
        lambda image, normalized_platform: VLLM_SR_DOCKER_IMAGE_DEFAULT,
    )
    monkeypatch.setattr(
        docker_images,
        "_ensure_image_available",
        lambda image, pull_policy: ensured.append((image, pull_policy)),
    )

    images = docker_images.get_runtime_images(pull_policy="never")

    assert images == {
        "router": VLLM_SR_DOCKER_IMAGE_DEFAULT,
        "envoy": VLLM_SR_ENVOY_DOCKER_IMAGE_DEFAULT,
        "dashboard": VLLM_SR_DASHBOARD_DOCKER_IMAGE_DEFAULT,
    }
    assert ensured == [
        (VLLM_SR_DOCKER_IMAGE_DEFAULT, "never"),
        (VLLM_SR_ENVOY_DOCKER_IMAGE_DEFAULT, "never"),
        (VLLM_SR_DASHBOARD_DOCKER_IMAGE_DEFAULT, "never"),
    ]


def test_get_runtime_images_derives_official_envoy_image_from_rocm_base(monkeypatch):
    ensured = []
    monkeypatch.setattr(
        docker_images,
        "_resolve_selected_image",
        lambda image, normalized_platform: VLLM_SR_DOCKER_IMAGE_ROCM,
    )
    monkeypatch.setattr(
        docker_images,
        "_ensure_image_available",
        lambda image, pull_policy: ensured.append((image, pull_policy)),
    )

    images = docker_images.get_runtime_images(pull_policy="never", platform="amd")

    assert images == {
        "router": VLLM_SR_DOCKER_IMAGE_ROCM,
        "envoy": VLLM_SR_ENVOY_DOCKER_IMAGE_DEFAULT,
        "dashboard": VLLM_SR_DASHBOARD_DOCKER_IMAGE_DEFAULT,
    }
    assert ensured == [
        (VLLM_SR_DOCKER_IMAGE_ROCM, "never"),
        (VLLM_SR_ENVOY_DOCKER_IMAGE_DEFAULT, "never"),
        (VLLM_SR_DASHBOARD_DOCKER_IMAGE_DEFAULT, "never"),
    ]


def test_get_runtime_images_reuses_official_router_image_from_official_tag(
    monkeypatch,
):
    ensured = []
    monkeypatch.setattr(
        docker_images,
        "_resolve_selected_image",
        lambda image, normalized_platform: "ghcr.io/vllm-project/semantic-router/vllm-sr:v1.2.3",
    )
    monkeypatch.setattr(
        docker_images,
        "_ensure_image_available",
        lambda image, pull_policy: ensured.append((image, pull_policy)),
    )

    images = docker_images.get_runtime_images(pull_policy="never")

    assert images["router"] == "ghcr.io/vllm-project/semantic-router/vllm-sr:v1.2.3"
    assert ensured == [
        ("ghcr.io/vllm-project/semantic-router/vllm-sr:v1.2.3", "never"),
        (VLLM_SR_ENVOY_DOCKER_IMAGE_DEFAULT, "never"),
        ("ghcr.io/vllm-project/semantic-router/dashboard:v1.2.3", "never"),
    ]


def test_get_runtime_images_derives_official_dashboard_image_from_official_tag(
    monkeypatch,
):
    ensured = []
    monkeypatch.setattr(
        docker_images,
        "_resolve_selected_image",
        lambda image, normalized_platform: "ghcr.io/vllm-project/semantic-router/vllm-sr:v1.2.3",
    )
    monkeypatch.setattr(
        docker_images,
        "_ensure_image_available",
        lambda image, pull_policy: ensured.append((image, pull_policy)),
    )

    images = docker_images.get_runtime_images(pull_policy="never")

    assert (
        images["dashboard"] == "ghcr.io/vllm-project/semantic-router/dashboard:v1.2.3"
    )
    assert ensured == [
        ("ghcr.io/vllm-project/semantic-router/vllm-sr:v1.2.3", "never"),
        (VLLM_SR_ENVOY_DOCKER_IMAGE_DEFAULT, "never"),
        ("ghcr.io/vllm-project/semantic-router/dashboard:v1.2.3", "never"),
    ]


def test_get_runtime_images_prefers_explicit_base_image_only_for_router(monkeypatch):
    ensured = []
    monkeypatch.setattr(
        docker_images,
        "_resolve_selected_image",
        lambda image, normalized_platform: image,
    )
    monkeypatch.setattr(
        docker_images,
        "_ensure_image_available",
        lambda image, pull_policy: ensured.append((image, pull_policy)),
    )
    monkeypatch.setenv("VLLM_SR_ROUTER_IMAGE", "router:from-env")

    images = docker_images.get_runtime_images(
        image="base:explicit",
        pull_policy="never",
    )

    assert images == {
        "router": "base:explicit",
        "envoy": VLLM_SR_ENVOY_DOCKER_IMAGE_DEFAULT,
        "dashboard": VLLM_SR_DASHBOARD_DOCKER_IMAGE_DEFAULT,
    }
    assert ensured == [
        ("base:explicit", "never"),
        (VLLM_SR_ENVOY_DOCKER_IMAGE_DEFAULT, "never"),
        (VLLM_SR_DASHBOARD_DOCKER_IMAGE_DEFAULT, "never"),
    ]


def test_get_runtime_images_prefers_service_env_over_explicit_base_for_nonrouter(
    monkeypatch,
):
    ensured = []
    monkeypatch.setattr(
        docker_images,
        "_resolve_selected_image",
        lambda image, normalized_platform: image,
    )
    monkeypatch.setattr(
        docker_images,
        "_ensure_image_available",
        lambda image, pull_policy: ensured.append((image, pull_policy)),
    )
    monkeypatch.setenv("VLLM_SR_ROUTER_IMAGE", "router:from-env")
    monkeypatch.setenv("VLLM_SR_ENVOY_IMAGE", "envoy:from-env")
    monkeypatch.setenv("VLLM_SR_DASHBOARD_IMAGE", "dashboard:from-env")

    images = docker_images.get_runtime_images(
        image="base:explicit",
        pull_policy="never",
    )

    assert images == {
        "router": "base:explicit",
        "envoy": "envoy:from-env",
        "dashboard": "dashboard:from-env",
    }
    assert ensured == [
        ("base:explicit", "never"),
        ("envoy:from-env", "never"),
        ("dashboard:from-env", "never"),
    ]


def test_get_runtime_images_derives_official_service_images_from_explicit_base_image(
    monkeypatch,
):
    ensured = []
    explicit_base = "ghcr.io/vllm-project/semantic-router/vllm-sr:v9.9.9"
    monkeypatch.setattr(
        docker_images,
        "_resolve_selected_image",
        lambda image, normalized_platform: image,
    )
    monkeypatch.setattr(
        docker_images,
        "_ensure_image_available",
        lambda image, pull_policy: ensured.append((image, pull_policy)),
    )

    images = docker_images.get_runtime_images(
        image=explicit_base,
        pull_policy="never",
    )

    assert images == {
        "router": explicit_base,
        "envoy": VLLM_SR_ENVOY_DOCKER_IMAGE_DEFAULT,
        "dashboard": "ghcr.io/vllm-project/semantic-router/dashboard:v9.9.9",
    }
    assert ensured == [
        (explicit_base, "never"),
        (VLLM_SR_ENVOY_DOCKER_IMAGE_DEFAULT, "never"),
        ("ghcr.io/vllm-project/semantic-router/dashboard:v9.9.9", "never"),
    ]


def test_get_runtime_images_prefers_explicit_service_image_over_explicit_base_image(
    monkeypatch,
):
    ensured = []
    monkeypatch.setattr(
        docker_images,
        "_resolve_selected_image",
        lambda image, normalized_platform: image,
    )
    monkeypatch.setattr(
        docker_images,
        "_ensure_image_available",
        lambda image, pull_policy: ensured.append((image, pull_policy)),
    )

    images = docker_images.get_runtime_images(
        image="base:explicit",
        envoy_image="envoy:explicit",
        pull_policy="never",
    )

    assert images == {
        "router": "base:explicit",
        "envoy": "envoy:explicit",
        "dashboard": VLLM_SR_DASHBOARD_DOCKER_IMAGE_DEFAULT,
    }
    assert ensured == [
        ("base:explicit", "never"),
        ("envoy:explicit", "never"),
        (VLLM_SR_DASHBOARD_DOCKER_IMAGE_DEFAULT, "never"),
    ]


def test_get_runtime_images_upgrades_router_override_to_rocm_on_amd(monkeypatch):
    ensured = []
    monkeypatch.setattr(
        docker_images,
        "_resolve_selected_image",
        lambda image, normalized_platform: "base:latest",
    )
    monkeypatch.setattr(
        docker_images,
        "_ensure_image_available",
        lambda image, pull_policy: ensured.append((image, pull_policy)),
    )

    images = docker_images.get_runtime_images(
        router_image="ghcr.io/vllm-project/semantic-router/vllm-sr:custom",
        pull_policy="never",
        platform="amd",
    )

    assert (
        images["router"] == "ghcr.io/vllm-project/semantic-router/vllm-sr-rocm:custom"
    )
    assert images["envoy"] == "base:latest"
    assert images["dashboard"] == "base:latest"
    assert ensured == [
        ("ghcr.io/vllm-project/semantic-router/vllm-sr-rocm:custom", "never"),
        ("base:latest", "never"),
    ]


def test_get_runtime_images_leaves_nonofficial_router_override_unchanged_on_amd(
    monkeypatch,
):
    ensured = []
    monkeypatch.setattr(
        docker_images,
        "_resolve_selected_image",
        lambda image, normalized_platform: "base:latest",
    )
    monkeypatch.setattr(
        docker_images,
        "_ensure_image_available",
        lambda image, pull_policy: ensured.append((image, pull_policy)),
    )

    images = docker_images.get_runtime_images(
        router_image="router:custom",
        pull_policy="never",
        platform="amd",
    )

    assert images["router"] == "router:custom"
    assert images["envoy"] == "base:latest"
    assert images["dashboard"] == "base:latest"
    assert ensured == [
        ("router:custom", "never"),
        ("base:latest", "never"),
    ]



"""Kubernetes deployment backend — wraps Helm and kubectl operations."""

from __future__ import annotations

import os
import shutil
import subprocess
from typing import Any

from cli.config_translator import (
    load_profile_values,
    translate_config_to_helm_values,
    write_helm_values_file,
)
from cli.logo import print_vllm_logo
from cli.utils import get_logger

log = get_logger(__name__)

HELM_RELEASE_NAME = "semantic-router"
DEFAULT_NAMESPACE = "vllm-semantic-router-system"
CHART_REL_PATH = os.path.join("deploy", "helm", "semantic-router")
ENV_SECRET_NAME = "vllm-sr-env-secrets"

K8S_SERVICE_TO_LABEL: dict[str, str] = {
    "router": "app.kubernetes.io/name=semantic-router",
    "dashboard": "app.kubernetes.io/component=dashboard",
    "envoy": "app.kubernetes.io/component=envoy",
}


class K8sBackend:
    """DeploymentBackend implementation for Kubernetes via Helm."""

    def __init__(
        self,
        *,
        namespace: str | None = None,
        context: str | None = None,
        release_name: str | None = None,
        profile: str | None = None,
        chart_dir: str | None = None,
    ) -> None:
        self.namespace = namespace or DEFAULT_NAMESPACE
        self.context = context
        self.release_name = release_name or HELM_RELEASE_NAME
        self.profile = profile
        self.chart_dir = chart_dir or self._find_chart_dir()

    # -- DeploymentBackend interface ------------------------------------------

    def deploy(
        self,
        config_file: str,
        env_vars: dict[str, str] | None = None,
        *,
        image: str | None = None,
        pull_policy: str | None = None,
        enable_observability: bool = True,
        **kwargs: Any,
    ) -> None:
        self._require_tool("helm")
        self._require_tool("kubectl")

        print_vllm_logo()
        log.info("Deploying vLLM Semantic Router to Kubernetes")
        log.info(f"  Release:   {self.release_name}")
        log.info(f"  Namespace: {self.namespace}")
        log.info(f"  Chart:     {self.chart_dir}")
        if self.context:
            log.info(f"  Context:   {self.context}")

        secret_name = self._sync_env_secret(env_vars)

        profile_values = load_profile_values(self.profile, self.chart_dir)
        values = translate_config_to_helm_values(
            config_file,
            image=image,
            pull_policy=pull_policy,
            enable_observability=enable_observability,
            profile_values=profile_values,
            env_vars=env_vars,
            env_secret_name=secret_name,
        )
        values_path = write_helm_values_file(values)

        cmd = [
            *self._helm_base_cmd(),
            "upgrade",
            "--install",
            self.release_name,
            self.chart_dir,
            "--namespace",
            self.namespace,
            "--create-namespace",
            "-f",
            values_path,
            "--wait",
            "--timeout",
            "10m",
        ]

        log.info("Running helm upgrade --install ...")
        self._run(cmd)
        log.info("Helm release deployed successfully")

        self._wait_for_pods()
        self._log_k8s_summary()

    def teardown(self) -> None:
        self._require_tool("helm")

        log.info(f"Uninstalling Helm release: {self.release_name}")
        cmd = [
            *self._helm_base_cmd(),
            "uninstall",
            self.release_name,
            "--namespace",
            self.namespace,
        ]
        self._run(cmd, check=False)
        self._delete_secret_if_exists(ENV_SECRET_NAME)
        log.info("Helm release uninstalled")

    def logs(self, service: str, follow: bool = False) -> None:
        self._require_tool("kubectl")

        label = self._label_for_service(service)
        cmd = [
            *self._kubectl_base_cmd(),
            "logs",
            "-l",
            label,
            "--namespace",
            self.namespace,
            "--all-containers",
            "--tail=200",
        ]
        if follow:
            cmd.append("--follow")

        log.info(f"Streaming {service} logs from Kubernetes ...")
        try:
            subprocess.run(cmd, check=False)
        except KeyboardInterrupt:
            log.info("\nLog streaming stopped")

    def status(self, service: str = "all") -> None:
        self._require_tool("kubectl")

        log.info("=" * 60)
        log.info(f"Kubernetes deployment status  (namespace: {self.namespace})")
        log.info("")

        self._run_display(
            [
                *self._kubectl_base_cmd(),
                "get",
                "pods",
                "--namespace",
                self.namespace,
                "-o",
                "wide",
            ]
        )
        log.info("")

        helm_cmd = [
            *self._helm_base_cmd(),
            "status",
            self.release_name,
            "--namespace",
            self.namespace,
            "--show-desc",
        ]
        self._run_display(helm_cmd)
        log.info("=" * 60)

    def get_dashboard_url(self) -> str | None:
        cmd = [
            *self._kubectl_base_cmd(),
            "get",
            "svc",
            "--namespace",
            self.namespace,
            "-l",
            "app.kubernetes.io/component=dashboard",
            "-o",
            "jsonpath={.items[0].spec.clusterIP}:{.items[0].spec.ports[0].port}",
        ]
        result = subprocess.run(cmd, capture_output=True, text=True, check=False)
        if result.returncode == 0 and result.stdout.strip():
            return f"http://{result.stdout.strip()}"
        return None

    def is_running(self) -> bool:
        cmd = [
            *self._helm_base_cmd(),
            "status",
            self.release_name,
            "--namespace",
            self.namespace,
            "-o",
            "json",
        ]
        result = subprocess.run(cmd, capture_output=True, text=True, check=False)
        return result.returncode == 0

    # -- helpers --------------------------------------------------------------

    def _sync_env_secret(self, env_vars: dict[str, str] | None) -> str | None:
        """Create or update a K8s Secret with sensitive env vars; return the secret name or None."""
        if not env_vars:
            return None

        from cli.commands.runtime_support import PASSTHROUGH_ENV_RULES  # noqa: PLC0415

        sensitive_names = {name for name, masked in PASSTHROUGH_ENV_RULES if masked}
        secret_data = {k: v for k, v in env_vars.items() if k in sensitive_names and v}

        if not secret_data:
            return None

        self._ensure_namespace()
        self._delete_secret_if_exists(ENV_SECRET_NAME)

        cmd = [
            *self._kubectl_base_cmd(),
            "create",
            "secret",
            "generic",
            ENV_SECRET_NAME,
            "--namespace",
            self.namespace,
        ]
        for key, value in sorted(secret_data.items()):
            cmd.append(f"--from-literal={key}={value}")

        log.info(
            f"Creating K8s secret '{ENV_SECRET_NAME}' with {len(secret_data)} key(s)"
        )
        self._run(cmd)
        return ENV_SECRET_NAME

    def _ensure_namespace(self) -> None:
        cmd = [
            *self._kubectl_base_cmd(),
            "create",
            "namespace",
            self.namespace,
            "--dry-run=client",
            "-o",
            "yaml",
        ]
        pipe_cmd = [*self._kubectl_base_cmd(), "apply", "-f", "-"]
        log.debug(f"Ensuring namespace {self.namespace}")
        ns_yaml = subprocess.run(cmd, capture_output=True, text=True, check=True)
        subprocess.run(pipe_cmd, input=ns_yaml.stdout, text=True, check=True)

    def _delete_secret_if_exists(self, name: str) -> None:
        cmd = [
            *self._kubectl_base_cmd(),
            "delete",
            "secret",
            name,
            "--namespace",
            self.namespace,
            "--ignore-not-found",
        ]
        self._run(cmd, check=False)

    def _helm_base_cmd(self) -> list[str]:
        cmd = ["helm"]
        if self.context:
            cmd += ["--kube-context", self.context]
        return cmd

    def _kubectl_base_cmd(self) -> list[str]:
        cmd = ["kubectl"]
        if self.context:
            cmd += ["--context", self.context]
        return cmd

    def _label_for_service(self, service: str) -> str:
        if service == "all":
            return f"app.kubernetes.io/instance={self.release_name}"
        label = K8S_SERVICE_TO_LABEL.get(service)
        if label is None:
            log.warning(
                f"Unknown service '{service}', falling back to release selector"
            )
            return f"app.kubernetes.io/instance={self.release_name}"
        return label

    def _wait_for_pods(self) -> None:
        log.info("Waiting for pods to become ready ...")
        cmd = [
            *self._kubectl_base_cmd(),
            "wait",
            "--for=condition=ready",
            "pod",
            "-l",
            f"app.kubernetes.io/instance={self.release_name}",
            "--namespace",
            self.namespace,
            "--timeout=600s",
        ]
        self._run(cmd, check=False)

    def _log_k8s_summary(self) -> None:
        log.info("=" * 60)
        log.info("vLLM Semantic Router deployed on Kubernetes!")
        log.info("")
        log.info("Commands:")
        log.info("  - vllm-sr status --target k8s")
        log.info("  - vllm-sr logs router --target k8s [-f]")
        log.info("  - vllm-sr stop --target k8s")
        log.info("")
        log.info("Port forwarding (access locally):")
        log.info(
            f"  kubectl port-forward -n {self.namespace} "
            f"svc/{self.release_name} 8080:8080"
        )
        log.info("=" * 60)

    @staticmethod
    def _require_tool(name: str) -> None:
        if shutil.which(name) is None:
            raise SystemExit(
                f"'{name}' is required for Kubernetes deployment but was not "
                "found on PATH."
            )

    @staticmethod
    def _find_chart_dir() -> str:
        candidates = [
            CHART_REL_PATH,
            os.path.join(os.getcwd(), CHART_REL_PATH),
        ]
        for path in candidates:
            if os.path.isdir(path) and os.path.exists(os.path.join(path, "Chart.yaml")):
                return os.path.abspath(path)
        raise SystemExit(
            f"Helm chart directory not found. Looked in: {candidates}. "
            "Set --chart-dir or run from the repository root."
        )

    @staticmethod
    def _run(cmd: list[str], check: bool = True) -> subprocess.CompletedProcess:
        log.debug(f"Running: {' '.join(cmd)}")
        return subprocess.run(cmd, check=check)

    @staticmethod
    def _run_display(cmd: list[str]) -> None:
        subprocess.run(cmd, check=False)



"""Shared helpers for Click command entrypoints."""

from __future__ import annotations

import sys
from collections.abc import Callable
from functools import wraps
from typing import Any


def exit_with_logged_error(
    log: Any, interrupt_message: str | None = None
) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """Wrap a CLI command and exit cleanly on logged failures."""

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            try:
                return func(*args, **kwargs)
            except KeyboardInterrupt:
                if interrupt_message:
                    log.info(interrupt_message)
                sys.exit(0)
            except Exception as exc:
                log.error(f"Error: {exc}")
                sys.exit(1)

        return wrapper

    return decorator



from pathlib import Path

REPO_ROOT = Path(__file__).resolve().parents[3]
INSTALL_SCRIPT_PATH = REPO_ROOT / "install.sh"
INSTALL_DOC_PATH = REPO_ROOT / "website" / "docs" / "installation" / "installation.md"
PYPI_PUBLISH_WORKFLOW_PATH = REPO_ROOT / ".github" / "workflows" / "pypi-publish.yml"
ROOT_MAKEFILE_PATH = REPO_ROOT / "Makefile"
RELEASE_MAKEFILE_PATH = REPO_ROOT / "tools" / "make" / "release.mk"


def test_install_script_only_mentions_docker_runtime() -> None:
    content = INSTALL_SCRIPT_PATH.read_text(encoding="utf-8")

    assert "podman" not in content.lower()
    assert "--runtime auto|docker|skip" in content
    assert "Linux auto -> docker" in content


def test_installation_doc_only_mentions_docker_runtime() -> None:
    content = INSTALL_DOC_PATH.read_text(encoding="utf-8")

    assert "Podman" not in content
    assert "Docker" in content


def test_install_script_defaults_to_dev_channel() -> None:
    content = INSTALL_SCRIPT_PATH.read_text(encoding="utf-8")

    assert 'REQUESTED_CHANNEL="${VLLM_SR_INSTALL_CHANNEL:-dev}"' in content
    assert "--channel stable|dev" in content
    assert (
        '"$INSTALL_ROOT/venv/bin/python" -m pip install --disable-pip-version-check --upgrade --quiet --pre vllm-sr'
        in content
    )


def test_installation_doc_mentions_dev_default_and_stable_override() -> None:
    content = INSTALL_DOC_PATH.read_text(encoding="utf-8")

    assert "latest development `vllm-sr` release" in content
    assert "bash -s -- --channel stable" in content
    assert "pip install --pre vllm-sr" in content


def test_pypi_publish_workflow_does_not_push_back_to_main() -> None:
    content = PYPI_PUBLISH_WORKFLOW_PATH.read_text(encoding="utf-8")

    assert "Bump next development base version on main" not in content
    assert "git push origin HEAD:main" not in content


def test_make_release_target_is_available_from_repo_root() -> None:
    root_makefile = ROOT_MAKEFILE_PATH.read_text(encoding="utf-8")
    release_makefile = RELEASE_MAKEFILE_PATH.read_text(encoding="utf-8")

    assert "tools/make/release.mk" in root_makefile
    assert "release:" in release_makefile
    assert (
        'src/vllm-sr/scripts/release.sh "$(RELEASE_VERSION)" "$(NEXT_VERSION)"'
        in release_makefile
    )



from pathlib import Path

REPO_ROOT = Path(__file__).resolve().parents[3]
DASHBOARD_DOCKERFILE = REPO_ROOT / "dashboard" / "backend" / "Dockerfile"
VLLM_SR_DOCKERFILE = REPO_ROOT / "src" / "vllm-sr" / "Dockerfile"
VLLM_SR_ROCM_DOCKERFILE = REPO_ROOT / "src" / "vllm-sr" / "Dockerfile.rocm"


def test_dashboard_dockerfile_uses_glibc_builder_for_cgo_backend() -> None:
    content = DASHBOARD_DOCKERFILE.read_text(encoding="utf-8")

    assert "FROM golang:1.24-bookworm AS backend-builder" in content
    assert "apt_get_install_with_retry build-essential" in content


def test_dashboard_dockerfile_retries_runtime_apk_installs() -> None:
    content = DASHBOARD_DOCKERFILE.read_text(encoding="utf-8")

    assert "FROM python:3.11-slim-bookworm" in content
    assert (
        "apt_get_install_with_retry ca-certificates curl docker.io gosu wget" in content
    )


def test_dashboard_dockerfile_copies_router_dsl_package_for_backend_builds() -> None:
    content = DASHBOARD_DOCKERFILE.read_text(encoding="utf-8")

    assert (
        "COPY src/semantic-router/pkg/dsl/ /app/src/semantic-router/pkg/dsl/" in content
    )
    assert (
        "COPY src/semantic-router/pkg/nlgen/ /app/src/semantic-router/pkg/nlgen/"
        in content
    )
    assert (
        "COPY src/semantic-router/internal/nlgen/ /app/src/semantic-router/internal/nlgen/"
        in content
    )


def test_dashboard_dockerfile_copies_model_eval_scripts_and_requirements() -> None:
    content = DASHBOARD_DOCKERFILE.read_text(encoding="utf-8")

    assert "COPY src/training/model_eval/ /app/src/training/model_eval/" in content
    assert (
        '"${VIRTUAL_ENV}/bin/pip" install --no-cache-dir -r /app/src/training/model_eval/requirements.txt'
        in content
    )


def test_vllm_sr_dockerfile_copies_router_dsl_package_for_dashboard_builder() -> None:
    content = VLLM_SR_DOCKERFILE.read_text(encoding="utf-8")

    assert (
        "COPY src/semantic-router/pkg/dsl/ /app/src/semantic-router/pkg/dsl/" in content
    )
    assert (
        "COPY src/semantic-router/pkg/nlgen/ /app/src/semantic-router/pkg/nlgen/"
        in content
    )
    assert (
        "COPY src/semantic-router/internal/nlgen/ /app/src/semantic-router/internal/nlgen/"
        in content
    )


def test_vllm_sr_rocm_dockerfile_copies_router_dsl_package_for_dashboard_builder() -> (
    None
):
    content = VLLM_SR_ROCM_DOCKERFILE.read_text(encoding="utf-8")

    assert (
        "COPY src/semantic-router/pkg/dsl/ /app/src/semantic-router/pkg/dsl/" in content
    )
    assert (
        "COPY src/semantic-router/pkg/nlgen/ /app/src/semantic-router/pkg/nlgen/"
        in content
    )
    assert (
        "COPY src/semantic-router/internal/nlgen/ /app/src/semantic-router/internal/nlgen/"
        in content
    )



from pathlib import Path

import yaml
from cli.bootstrap import build_bootstrap_config
from cli.commands.runtime_support import (
    append_passthrough_env_vars,
    apply_runtime_mode_env_vars,
    configure_runtime_override_env_vars,
    resolve_effective_config_path,
)

REPO_ROOT = Path(__file__).resolve().parents[3]


def test_apply_runtime_mode_env_vars_sets_dashboard_readonly_when_requested():
    env_vars: dict[str, str] = {}

    apply_runtime_mode_env_vars(
        env_vars=env_vars,
        minimal=False,
        readonly=True,
        setup_mode=False,
        platform=None,
    )

    assert env_vars["DASHBOARD_READONLY"] == "true"


def test_apply_runtime_mode_env_vars_skips_dashboard_readonly_in_minimal_mode():
    env_vars: dict[str, str] = {}

    apply_runtime_mode_env_vars(
        env_vars=env_vars,
        minimal=True,
        readonly=True,
        setup_mode=False,
        platform=None,
    )

    assert env_vars["DISABLE_DASHBOARD"] == "true"
    assert "DASHBOARD_READONLY" not in env_vars


def test_apply_runtime_mode_env_vars_sets_router_log_level_when_requested():
    env_vars: dict[str, str] = {}

    apply_runtime_mode_env_vars(
        env_vars=env_vars,
        minimal=False,
        readonly=False,
        setup_mode=False,
        platform=None,
        log_level="DEBUG",
    )

    assert env_vars["SR_LOG_LEVEL"] == "debug"


def test_append_passthrough_env_vars_includes_router_logging_settings(monkeypatch):
    monkeypatch.setenv("SR_LOG_LEVEL", "debug")
    monkeypatch.setenv("SR_LOG_ENCODING", "console")

    env_vars: dict[str, str] = {}
    append_passthrough_env_vars(env_vars)

    assert env_vars["SR_LOG_LEVEL"] == "debug"
    assert env_vars["SR_LOG_ENCODING"] == "console"


def test_resolve_effective_config_path_applies_amd_gpu_defaults(tmp_path: Path):
    config_path = tmp_path / "config.yaml"
    config_path.write_text(
        yaml.safe_dump(
            {
                "version": "v0.3",
                "global": {
                    "model_catalog": {
                        "embeddings": {
                            "semantic": {
                                "use_cpu": True,
                                "embedding_config": {"model_type": "mmbert"},
                            }
                        },
                        "modules": {
                            "prompt_guard": {"use_cpu": True},
                            "classifier": {
                                "domain": {"use_cpu": True},
                            },
                        },
                    }
                },
            },
            sort_keys=False,
        )
    )

    effective_path = resolve_effective_config_path(
        config_path=config_path,
        algorithm=None,
        setup_mode=False,
        platform="amd",
    )

    assert effective_path != config_path
    assert effective_path == tmp_path / ".vllm-sr" / "runtime-config.yaml"
    effective = yaml.safe_load(effective_path.read_text())
    assert (
        effective["global"]["model_catalog"]["embeddings"]["semantic"]["use_cpu"]
        is False
    )
    assert (
        effective["global"]["model_catalog"]["modules"]["prompt_guard"]["use_cpu"]
        is False
    )
    assert (
        effective["global"]["model_catalog"]["modules"]["classifier"]["domain"][
            "use_cpu"
        ]
        is False
    )


def test_resolve_effective_config_path_combines_algorithm_and_platform_overrides(
    tmp_path: Path,
):
    config_path = tmp_path / "config.yaml"
    config_path.write_text(
        yaml.safe_dump(
            {
                "version": "v0.3",
                "routing": {"decisions": [{"name": "default"}]},
                "global": {
                    "model_catalog": {
                        "embeddings": {
                            "semantic": {
                                "use_cpu": True,
                                "embedding_config": {"model_type": "mmbert"},
                            }
                        }
                    }
                },
            },
            sort_keys=False,
        )
    )

    effective_path = resolve_effective_config_path(
        config_path=config_path,
        algorithm="elo",
        setup_mode=False,
        platform="amd",
    )

    assert effective_path == tmp_path / ".vllm-sr" / "runtime-config.yaml"
    effective = yaml.safe_load(effective_path.read_text())
    assert effective["routing"]["decisions"][0]["algorithm"]["type"] == "elo"
    assert (
        effective["global"]["model_catalog"]["embeddings"]["semantic"]["use_cpu"]
        is False
    )


def test_resolve_effective_config_path_injects_missing_amd_gpu_defaults(
    tmp_path: Path,
):
    config_path = tmp_path / "config.yaml"
    config_path.write_text(
        yaml.safe_dump(
            {
                "version": "v0.3",
                "listeners": [
                    {
                        "name": "http",
                        "address": "0.0.0.0",
                        "port": 8899,
                    }
                ],
                "providers": {
                    "defaults": {"default_model": "test-model"},
                    "models": [
                        {
                            "name": "test-model",
                            "provider_model_id": "test-model",
                            "backend_refs": [{"endpoint": "127.0.0.1:8000"}],
                        }
                    ],
                },
                "routing": {
                    "modelCards": [{"name": "test-model"}],
                    "decisions": [
                        {
                            "name": "default-route",
                            "priority": 1,
                            "rules": {"operator": "AND", "conditions": []},
                            "modelRefs": [{"model": "test-model"}],
                        }
                    ],
                },
                "global": {},
            },
            sort_keys=False,
        )
    )

    effective_path = resolve_effective_config_path(
        config_path=config_path,
        algorithm=None,
        setup_mode=False,
        platform="amd",
    )

    effective = yaml.safe_load(effective_path.read_text())
    model_catalog = effective["global"]["model_catalog"]
    assert model_catalog["embeddings"]["semantic"]["use_cpu"] is False
    assert model_catalog["modules"]["prompt_guard"]["use_cpu"] is False
    assert model_catalog["modules"]["classifier"]["domain"]["use_cpu"] is False
    assert model_catalog["modules"]["classifier"]["pii"]["use_cpu"] is False
    assert model_catalog["modules"]["feedback_detector"]["use_cpu"] is False
    assert (
        model_catalog["modules"]["modality_detector"]["classifier"]["use_cpu"] is False
    )
    assert "bert" not in model_catalog["embeddings"]


def test_resolve_effective_config_path_does_not_reintroduce_deprecated_bert_for_amd(
    tmp_path: Path,
):
    config_path = tmp_path / "config.yaml"
    balance_recipe = REPO_ROOT / "deploy" / "recipes" / "balance.yaml"
    config_path.write_text(balance_recipe.read_text(encoding="utf-8"))

    effective_path = resolve_effective_config_path(
        config_path=config_path,
        algorithm=None,
        setup_mode=False,
        platform="amd",
    )

    effective = yaml.safe_load(effective_path.read_text())
    model_catalog = effective.get("global", {}).get("model_catalog", {})
    embeddings = model_catalog.get("embeddings", {})
    assert "bert" not in embeddings
    assert embeddings["semantic"]["use_cpu"] is False


def test_configure_runtime_override_env_vars_sets_internal_runtime_path(tmp_path: Path):
    env_vars: dict[str, str] = {}
    source_config = tmp_path / "config.yaml"
    source_config.write_text("version: v0.3\n")
    effective_config = tmp_path / ".vllm-sr" / "runtime-config.yaml"
    effective_config.parent.mkdir(parents=True, exist_ok=True)
    effective_config.write_text("version: v0.3\n")

    configure_runtime_override_env_vars(env_vars, source_config, effective_config)

    assert env_vars["VLLM_SR_SOURCE_CONFIG_PATH"] == "/app/config.yaml"
    assert (
        env_vars["VLLM_SR_RUNTIME_CONFIG_PATH"] == "/app/.vllm-sr/runtime-config.yaml"
    )


def test_resolve_effective_config_path_stages_relative_kb_assets(tmp_path: Path):
    kb_root = tmp_path / "knowledge_bases"
    kb_root.mkdir(parents=True)
    (kb_root / "labels.json").write_text(
        '{"labels":{"safe":{"exemplars":["hello"]}}}', encoding="utf-8"
    )

    config_path = tmp_path / "config.yaml"
    config_path.write_text(
        yaml.safe_dump(
            {
                "version": "v0.3",
                "global": {
                    "model_catalog": {
                        "kbs": [
                            {
                                "name": "privacy_kb",
                                "source": {
                                    "path": "knowledge_bases/",
                                    "manifest": "labels.json",
                                },
                            }
                        ]
                    }
                },
            },
            sort_keys=False,
        )
    )

    effective_path = resolve_effective_config_path(
        config_path=config_path,
        algorithm=None,
        setup_mode=False,
        platform=None,
    )

    assert effective_path == tmp_path / ".vllm-sr" / "runtime-config.yaml"
    effective = yaml.safe_load(effective_path.read_text())
    assert (
        effective["global"]["model_catalog"]["kbs"][0]["source"]["path"]
        == "knowledge_bases/privacy_kb"
    )
    assert (
        effective_path.parent / "knowledge_bases" / "privacy_kb" / "labels.json"
    ).exists()


def test_resolve_effective_config_path_stages_builtin_kb_assets(tmp_path: Path):
    builtin_manifest = REPO_ROOT / "config" / "kb" / "privacy" / "labels.json"
    assert builtin_manifest.exists()

    config_path = tmp_path / "config.yaml"
    config_path.write_text(
        yaml.safe_dump(
            {
                "version": "v0.3",
                "global": {
                    "model_catalog": {
                        "kbs": [
                            {
                                "name": "privacy_kb",
                                "source": {
                                    "path": "kb/privacy/",
                                    "manifest": "labels.json",
                                },
                            }
                        ]
                    }
                },
            },
            sort_keys=False,
        )
    )

    effective_path = resolve_effective_config_path(
        config_path=config_path,
        algorithm=None,
        setup_mode=False,
        platform=None,
    )

    assert effective_path == tmp_path / ".vllm-sr" / "runtime-config.yaml"
    staged_manifest = (
        effective_path.parent / "knowledge_bases" / "privacy_kb" / "labels.json"
    )
    assert staged_manifest.exists()
    effective = yaml.safe_load(effective_path.read_text())
    assert (
        effective["global"]["model_catalog"]["kbs"][0]["source"]["path"]
        == "knowledge_bases/privacy_kb"
    )


def test_resolve_effective_config_path_injects_local_service_runtime_defaults(
    tmp_path: Path,
):
    config_path = tmp_path / "config.yaml"
    config_path.write_text(
        yaml.safe_dump(
            {
                "version": "v0.3",
                "listeners": [
                    {
                        "name": "http-8899",
                        "address": "0.0.0.0",
                        "port": 8899,
                    }
                ],
            },
            sort_keys=False,
        )
    )

    effective_path = resolve_effective_config_path(
        config_path=config_path,
        algorithm=None,
        setup_mode=False,
        platform=None,
    )

    assert effective_path == tmp_path / ".vllm-sr" / "runtime-config.yaml"
    effective = yaml.safe_load(effective_path.read_text())
    response_api = effective["global"]["services"]["response_api"]
    assert response_api["enabled"] is True
    assert response_api["store_backend"] == "redis"
    assert response_api["redis"]["address"] == "vllm-sr-redis:6379"
    assert response_api["redis"]["db"] == 0

    router_replay = effective["global"]["services"]["router_replay"]
    assert router_replay["enabled"] is True
    assert router_replay["store_backend"] == "postgres"
    assert router_replay["postgres"]["host"] == "vllm-sr-postgres"
    assert router_replay["postgres"]["port"] == 5432
    assert router_replay["postgres"]["database"] == "vsr"
    assert router_replay["postgres"]["user"] == "router"
    assert router_replay["postgres"]["password"] == "router-secret"
    assert router_replay["postgres"]["ssl_mode"] == "disable"


def test_resolve_effective_config_path_preserves_setup_mode_bootstrap_config(
    tmp_path: Path,
):
    config_path = tmp_path / "config.yaml"
    config_path.write_text(yaml.safe_dump(build_bootstrap_config(), sort_keys=False))

    effective_path = resolve_effective_config_path(
        config_path=config_path,
        algorithm=None,
        setup_mode=True,
        platform=None,
    )

    assert effective_path == config_path



"""Helpers for importing external config sources into canonical VSR config."""

from __future__ import annotations

import json
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import yaml
from pydantic import ValidationError as PydanticValidationError

from cli.config_migration import migrate_config_data
from cli.consts import DEFAULT_LISTENER_PORT
from cli.models import UserConfig
from cli.parser import ConfigParseError, load_config_file
from cli.utils import get_logger
from cli.validator import validate_user_config

log = get_logger(__name__)

OPENCLAW_CONFIG_ENV = "OPENCLAW_CONFIG_PATH"
SUPPORTED_OPENCLAW_API_PREFIXES = ("openai",)


class ConfigImportError(RuntimeError):
    """Raised when an import source cannot be converted safely."""


@dataclass(frozen=True)
class ImportedModel:
    """Imported OpenClaw model metadata and source payload."""

    provider_key: str
    source_model_id: str
    logical_name: str
    provider_config: dict[str, Any]
    model_config: dict[str, Any]


@dataclass(frozen=True)
class ImportResult:
    """Result of importing OpenClaw config into canonical VSR config."""

    source_path: Path
    source_backup_path: Path
    target_path: Path
    target_backup_path: Path | None
    rewritten_base_url: str
    imported_models: list[ImportedModel]


def import_config_command(
    from_type: str,
    source_path: str | None = None,
    target_path: str = "config.yaml",
    force: bool = False,
) -> ImportResult:
    """Import an external config source into canonical VSR config."""

    normalized_from = (from_type or "").strip().lower()
    if normalized_from != "openclaw":
        raise ConfigImportError(
            f"Unsupported import source '{from_type}'. Supported values: openclaw."
        )

    resolved_source = discover_openclaw_config(source_path)
    source_raw, source_data = load_openclaw_source(resolved_source)
    imported_models = collect_openclaw_models(source_data)
    resolved_target = Path(target_path).expanduser()
    target = load_or_bootstrap_target_config(resolved_target)

    merge_openclaw_models_into_target(target, imported_models)
    rewritten_base_url = build_listener_base_url(target)
    rewrite_openclaw_source(source_data, imported_models, rewritten_base_url)
    validate_import_result(target)

    resolved_target.parent.mkdir(parents=True, exist_ok=True)

    target_backup_path = (
        backup_path_for(resolved_target) if resolved_target.exists() else None
    )
    source_backup_path = backup_path_for(resolved_source)
    ensure_backup_paths_available(
        target_backup_path=target_backup_path,
        source_backup_path=source_backup_path,
        force=force,
    )

    if target_backup_path is not None:
        target_backup_path.write_text(
            resolved_target.read_text(encoding="utf-8"),
            encoding="utf-8",
        )

    resolved_target.write_text(
        yaml.safe_dump(target, sort_keys=False, allow_unicode=True),
        encoding="utf-8",
    )

    source_backup_path.write_text(source_raw, encoding="utf-8")
    resolved_source.write_text(
        json.dumps(source_data, indent=2, ensure_ascii=False) + "\n",
        encoding="utf-8",
    )

    log.info("Imported OpenClaw configuration written successfully")
    log.info(f"  Source: {resolved_source}")
    log.info(f"  Source backup: {source_backup_path}")
    log.info(f"  Target: {resolved_target}")
    if target_backup_path is not None:
        log.info(f"  Target backup: {target_backup_path}")
    log.info(f"  Rewritten OpenClaw base URL: {rewritten_base_url}")
    log.info(
        "  Imported models: %s",
        ", ".join(model.logical_name for model in imported_models),
    )

    return ImportResult(
        source_path=resolved_source,
        source_backup_path=source_backup_path,
        target_path=resolved_target,
        target_backup_path=target_backup_path,
        rewritten_base_url=rewritten_base_url,
        imported_models=imported_models,
    )


def discover_openclaw_config(source_path: str | None = None) -> Path:
    """Resolve the OpenClaw config path from an explicit path or discovery order."""

    if source_path:
        candidate = Path(source_path).expanduser()
        if not candidate.exists():
            raise ConfigImportError(f"OpenClaw config not found: {candidate}")
        if candidate.is_dir():
            raise ConfigImportError(f"OpenClaw config path is a directory: {candidate}")
        return candidate

    candidates: list[Path] = []
    if raw_env_path := os.getenv(OPENCLAW_CONFIG_ENV):
        candidates.append(Path(raw_env_path).expanduser())
    candidates.extend(
        [
            Path.cwd() / "openclaw.json",
            Path.home() / ".openclaw" / "openclaw.json",
        ]
    )

    checked: list[str] = []
    for candidate in candidates:
        checked.append(str(candidate))
        if candidate.exists() and candidate.is_file():
            return candidate

    raise ConfigImportError(
        "Could not find an OpenClaw config. "
        f"Checked {', '.join(checked)}. Set {OPENCLAW_CONFIG_ENV} or pass --source."
    )


def load_openclaw_source(source_path: Path) -> tuple[str, dict[str, Any]]:
    """Load OpenClaw JSON and return both raw text and parsed mapping."""

    try:
        raw = source_path.read_text(encoding="utf-8")
    except OSError as exc:
        raise ConfigImportError(
            f"Failed to read OpenClaw config {source_path}: {exc}"
        ) from exc

    try:
        data = json.loads(raw)
    except json.JSONDecodeError as exc:
        raise ConfigImportError(
            f"OpenClaw config {source_path} is not valid JSON: {exc}"
        ) from exc

    if not isinstance(data, dict):
        raise ConfigImportError(
            f"OpenClaw config {source_path} must contain a JSON object at the top level."
        )

    return raw, data


def collect_openclaw_models(source_data: dict[str, Any]) -> list[ImportedModel]:
    """Collect supported OpenClaw provider/model bindings."""

    providers = openclaw_provider_block(source_data)
    raw_entries: list[tuple[str, dict[str, Any], dict[str, Any]]] = []
    unsupported: list[str] = []
    duplicate_provider_model_ids: list[str] = []

    for provider_key, provider_config in providers.items():
        provider_entries, provider_unsupported, provider_duplicates = (
            collect_provider_entries(provider_key, provider_config)
        )
        if provider_unsupported:
            unsupported.append(provider_unsupported)
            continue
        raw_entries.extend(provider_entries)
        duplicate_provider_model_ids.extend(provider_duplicates)

    if unsupported:
        raise ConfigImportError(
            "Unsupported OpenClaw provider API families: "
            f"{', '.join(unsupported)}. "
            "Supported providers must use an OpenAI-compatible api value."
        )

    if duplicate_provider_model_ids:
        raise ConfigImportError(
            "OpenClaw config contains duplicate model ids within a provider: "
            f"{', '.join(duplicate_provider_model_ids)}."
        )

    if not raw_entries:
        raise ConfigImportError(
            "OpenClaw config contains no importable provider models under models.providers."
        )

    model_id_counts: dict[str, int] = {}
    for _, _, model in raw_entries:
        model_id = str(model["id"]).strip()
        model_id_counts[model_id] = model_id_counts.get(model_id, 0) + 1

    imported_models: list[ImportedModel] = []
    for provider_key, provider_config, model in raw_entries:
        source_model_id = str(model["id"]).strip()
        logical_name = (
            source_model_id
            if model_id_counts[source_model_id] == 1
            else f"{provider_key}/{source_model_id}"
        )
        imported_models.append(
            ImportedModel(
                provider_key=provider_key,
                source_model_id=source_model_id,
                logical_name=logical_name,
                provider_config=provider_config,
                model_config=model,
            )
        )
    return imported_models


def load_or_bootstrap_target_config(target_path: Path) -> dict[str, Any]:
    """Load an existing target config or bootstrap a minimal canonical config."""

    if target_path.exists():
        if target_path.is_dir():
            raise ConfigImportError(f"Target config path is a directory: {target_path}")
        try:
            data = load_config_file(str(target_path))
        except ConfigParseError as exc:
            raise ConfigImportError(
                f"Failed to read target config {target_path}: {exc}"
            ) from exc
        target = migrate_config_data(data)
    else:
        target = build_minimal_target_config()

    if not isinstance(target, dict):
        raise ConfigImportError(
            f"Target config {target_path} did not resolve to a mapping."
        )

    normalize_target_shape(target)
    return target


def build_minimal_target_config() -> dict[str, Any]:
    """Build the minimal canonical config used when the target does not exist."""

    return {
        "version": "v0.3",
        "listeners": [default_listener()],
        "providers": {
            "defaults": {},
            "models": [],
        },
        "routing": {
            "modelCards": [],
            "signals": {},
            "decisions": [],
        },
    }


def default_listener() -> dict[str, Any]:
    """Return the default local listener used by bootstrapped imports."""

    return {
        "name": f"http-{DEFAULT_LISTENER_PORT}",
        "address": "0.0.0.0",
        "port": DEFAULT_LISTENER_PORT,
        "timeout": "300s",
    }


def merge_openclaw_models_into_target(
    target: dict[str, Any],
    imported_models: list[ImportedModel],
) -> None:
    """Merge imported models into the canonical target config."""

    providers = target["providers"]
    routing = target["routing"]
    provider_models = providers["models"]
    routing_cards = routing["modelCards"]
    defaults = providers["defaults"]

    provider_models_by_name = {
        str(model.get("name", "")).strip(): model
        for model in provider_models
        if isinstance(model, dict) and str(model.get("name", "")).strip()
    }
    routing_cards_by_name = {
        str(card.get("name", "")).strip(): card
        for card in routing_cards
        if isinstance(card, dict) and str(card.get("name", "")).strip()
    }

    for imported_model in imported_models:
        provider_model = provider_models_by_name.get(imported_model.logical_name)
        if provider_model is None:
            provider_model = {"name": imported_model.logical_name}
            provider_models.append(provider_model)
            provider_models_by_name[imported_model.logical_name] = provider_model

        provider_model["provider_model_id"] = imported_model.source_model_id
        provider_model["api_format"] = provider_model.get("api_format") or "openai"

        external_model_ids = provider_model.get("external_model_ids")
        if not isinstance(external_model_ids, dict):
            external_model_ids = {}
            provider_model["external_model_ids"] = external_model_ids
        external_model_ids["openai"] = imported_model.source_model_id

        provider_model["backend_refs"] = [build_backend_ref(imported_model)]

        routing_card = routing_cards_by_name.get(imported_model.logical_name)
        if routing_card is None:
            routing_card = {"name": imported_model.logical_name}
            routing_cards.append(routing_card)
            routing_cards_by_name[imported_model.logical_name] = routing_card

        routing_card["name"] = imported_model.logical_name
        context_window = positive_int(imported_model.model_config.get("contextWindow"))
        if context_window is not None:
            routing_card["context_window_size"] = context_window
        capabilities = build_capabilities(imported_model.model_config)
        if capabilities:
            routing_card["capabilities"] = capabilities
        routing_card.setdefault(
            "description",
            f"Imported from OpenClaw provider '{imported_model.provider_key}'.",
        )

    provider_names = [
        str(model.get("name", "")).strip()
        for model in provider_models
        if isinstance(model, dict) and str(model.get("name", "")).strip()
    ]
    default_model = str(defaults.get("default_model", "") or "").strip()
    if not default_model or default_model not in provider_names:
        defaults["default_model"] = imported_models[0].logical_name

    decisions = routing["decisions"]
    if not decisions:
        routing["decisions"] = [build_default_decision(defaults["default_model"])]


def build_backend_ref(imported_model: ImportedModel) -> dict[str, Any]:
    """Build one canonical backend ref from an OpenClaw provider definition."""

    backend_ref: dict[str, Any] = {
        "name": imported_model.provider_key,
        "base_url": str(
            imported_model.provider_config.get("baseUrl", "") or ""
        ).strip(),
        "provider": "openai",
        "weight": 1,
    }

    api_key = str(imported_model.provider_config.get("apiKey", "") or "").strip()
    if api_key and api_key != "not-needed":
        backend_ref["api_key"] = api_key
        backend_ref["auth_header"] = "Authorization"
        backend_ref["auth_prefix"] = "Bearer"

    headers = imported_model.provider_config.get("headers")
    if isinstance(headers, dict):
        extra_headers = {
            str(key): str(value)
            for key, value in headers.items()
            if value is not None and str(key)
        }
        if extra_headers:
            backend_ref["extra_headers"] = extra_headers

    return backend_ref


def build_capabilities(model_config: dict[str, Any]) -> list[str]:
    """Map OpenClaw model metadata into routing.modelCards capabilities."""

    capabilities: list[str] = []
    inputs = model_config.get("input")
    normalized_inputs = []
    if isinstance(inputs, list):
        normalized_inputs = [
            str(item).strip().lower() for item in inputs if str(item).strip()
        ]

    if not normalized_inputs or "text" in normalized_inputs:
        capabilities.append("chat")
    if "image" in normalized_inputs:
        capabilities.append("vision")
    if "audio" in normalized_inputs:
        capabilities.append("audio")
    if bool(model_config.get("reasoning")):
        capabilities.append("reasoning")

    seen: set[str] = set()
    unique_capabilities: list[str] = []
    for capability in capabilities:
        if capability in seen:
            continue
        seen.add(capability)
        unique_capabilities.append(capability)
    return unique_capabilities


def build_default_decision(default_model: str) -> dict[str, Any]:
    """Build the catch-all decision used for a bootstrapped target."""

    return {
        "name": "openclaw-import-default",
        "description": "Catch-all route bootstrapped from OpenClaw import.",
        "priority": 100,
        "rules": {
            "operator": "AND",
            "conditions": [],
        },
        "modelRefs": [
            {
                "model": default_model,
            }
        ],
    }


def build_listener_base_url(target: dict[str, Any]) -> str:
    """Resolve the first listener into a local OpenClaw-compatible base URL."""

    listeners = target.get("listeners")
    if not isinstance(listeners, list) or not listeners:
        raise ConfigImportError(
            "Imported config must declare at least one listener before OpenClaw can be rewritten."
        )

    first_listener = listeners[0]
    if not isinstance(first_listener, dict):
        raise ConfigImportError("The first listener must be a mapping.")

    port = positive_int(first_listener.get("port")) or DEFAULT_LISTENER_PORT
    raw_address = str(first_listener.get("address", "") or "").strip()
    host = "127.0.0.1" if raw_address in {"", "0.0.0.0", "::", "[::]"} else raw_address
    return f"http://{host}:{port}/v1"


def rewrite_openclaw_source(
    source_data: dict[str, Any],
    imported_models: list[ImportedModel],
    rewritten_base_url: str,
) -> None:
    """Rewrite imported OpenClaw provider base URLs and collision-renamed model ids."""

    providers = source_data["models"]["providers"]
    replacements: dict[str, str] = {}
    renamed_model_ids: list[tuple[str, str, str]] = []

    for imported_model in imported_models:
        provider_config = providers.get(imported_model.provider_key)
        if not isinstance(provider_config, dict):
            continue
        provider_config["baseUrl"] = rewritten_base_url
        collect_rewrite_changes(
            provider_config.get("models"),
            imported_model,
            replacements,
            renamed_model_ids,
        )

    if replacements:
        replace_openclaw_model_refs(source_data, replacements)

    apply_renamed_model_ids(providers, renamed_model_ids)


def replace_openclaw_model_refs(node: Any, replacements: dict[str, str]) -> Any:
    """Recursively replace OpenClaw provider/model reference strings."""

    if isinstance(node, dict):
        for key, value in list(node.items()):
            node[key] = replace_openclaw_model_refs(value, replacements)
        return node
    if isinstance(node, list):
        for index, value in enumerate(list(node)):
            node[index] = replace_openclaw_model_refs(value, replacements)
        return node
    if isinstance(node, str):
        return replacements.get(node, node)
    return node


def validate_import_result(target: dict[str, Any]) -> None:
    """Validate the merged canonical config before writing it to disk."""

    try:
        config = UserConfig(**target)
    except PydanticValidationError as exc:
        details = "; ".join(
            f"{'.'.join(str(part) for part in error['loc'])}: {error['msg']}"
            for error in exc.errors()
        )
        raise ConfigImportError(
            f"Imported config does not satisfy the canonical schema: {details}"
        ) from exc

    errors = validate_user_config(config)
    if errors:
        rendered = "; ".join(str(error) for error in errors)
        raise ConfigImportError(f"Imported config failed validation: {rendered}")


def backup_path_for(path: Path) -> Path:
    """Return the default backup path for an imported or rewritten file."""

    return path.with_name(f"{path.name}.bak")


def ensure_backup_paths_available(
    target_backup_path: Path | None,
    source_backup_path: Path,
    force: bool,
) -> None:
    """Fail before writing when backup files already exist and --force is absent."""

    existing: list[str] = []
    for path in (target_backup_path, source_backup_path):
        if path is not None and path.exists():
            existing.append(str(path))

    if existing and not force:
        raise ConfigImportError(
            "Backup file already exists: "
            f"{', '.join(existing)}. Use --force to overwrite backup files."
        )


def positive_int(value: Any) -> int | None:
    """Return a positive integer value or None."""

    if isinstance(value, bool):
        return None
    try:
        parsed = int(value)
    except (TypeError, ValueError):
        return None
    return parsed if parsed > 0 else None


def openclaw_provider_block(source_data: dict[str, Any]) -> dict[str, Any]:
    """Return the OpenClaw provider block or raise an actionable error."""

    models_block = source_data.get("models")
    providers = (
        models_block.get("providers") if isinstance(models_block, dict) else None
    )
    if not isinstance(providers, dict) or not providers:
        raise ConfigImportError(
            "OpenClaw config is missing models.providers. "
            "Supported imports require models.providers.* with OpenAI-compatible endpoints."
        )
    return providers


def collect_provider_entries(
    provider_key: str,
    provider_config: Any,
) -> tuple[list[tuple[str, dict[str, Any], dict[str, Any]]], str | None, list[str]]:
    """Collect valid model entries for one OpenClaw provider."""

    if not isinstance(provider_config, dict):
        raise ConfigImportError(
            f"OpenClaw provider '{provider_key}' must be a JSON object."
        )

    provider_api = str(provider_config.get("api", "") or "").strip().lower()
    if provider_api and not provider_api.startswith(SUPPORTED_OPENCLAW_API_PREFIXES):
        return [], f"{provider_key} ({provider_api})", []

    base_url = str(provider_config.get("baseUrl", "") or "").strip()
    if not base_url:
        raise ConfigImportError(
            f"OpenClaw provider '{provider_key}' is missing baseUrl."
        )

    models = provider_config.get("models")
    if models in (None, []):
        return [], None, []
    if not isinstance(models, list):
        raise ConfigImportError(
            f"OpenClaw provider '{provider_key}'.models must be an array."
        )

    entries: list[tuple[str, dict[str, Any], dict[str, Any]]] = []
    duplicates: list[str] = []
    seen_ids: set[str] = set()

    for model in models:
        model_id = validate_provider_model(provider_key, model)
        if model_id in seen_ids:
            duplicates.append(f"{provider_key}/{model_id}")
            continue
        seen_ids.add(model_id)
        entries.append((provider_key, provider_config, model))

    return entries, None, duplicates


def validate_provider_model(provider_key: str, model: Any) -> str:
    """Validate one OpenClaw provider model entry and return its id."""

    if not isinstance(model, dict):
        raise ConfigImportError(
            f"OpenClaw provider '{provider_key}' contains a non-object model entry."
        )

    model_id = str(model.get("id", "") or "").strip()
    if not model_id:
        raise ConfigImportError(
            f"OpenClaw provider '{provider_key}' contains a model without id."
        )
    return model_id


def normalize_target_shape(target: dict[str, Any]) -> None:
    """Normalize a loaded target into the canonical blocks required by import."""

    target["version"] = "v0.3"
    target.pop("setup", None)

    listeners = ensure_list(target, "listeners")
    if not listeners:
        target["listeners"] = [default_listener()]

    providers = ensure_mapping(target, "providers")
    ensure_mapping(providers, "defaults")
    ensure_list(providers, "models")

    routing = ensure_mapping(target, "routing")
    ensure_list(routing, "modelCards")
    ensure_mapping(routing, "signals")
    ensure_list(routing, "decisions")


def ensure_mapping(parent: dict[str, Any], key: str) -> dict[str, Any]:
    """Ensure a nested mapping key exists."""

    value = parent.get(key)
    if not isinstance(value, dict):
        value = {}
        parent[key] = value
    return value


def ensure_list(parent: dict[str, Any], key: str) -> list[Any]:
    """Ensure a nested list key exists."""

    value = parent.get(key)
    if not isinstance(value, list):
        value = []
        parent[key] = value
    return value


def collect_rewrite_changes(
    models: Any,
    imported_model: ImportedModel,
    replacements: dict[str, str],
    renamed_model_ids: list[tuple[str, str, str]],
) -> None:
    """Collect provider/model ref replacements and queued model-id renames."""

    if not isinstance(models, list):
        return

    for model in models:
        if (
            isinstance(model, dict)
            and str(model.get("id", "") or "").strip() == imported_model.source_model_id
        ):
            if imported_model.logical_name != imported_model.source_model_id:
                old_ref = (
                    f"{imported_model.provider_key}/{imported_model.source_model_id}"
                )
                new_ref = f"{imported_model.provider_key}/{imported_model.logical_name}"
                replacements[old_ref] = new_ref
                renamed_model_ids.append(
                    (
                        imported_model.provider_key,
                        imported_model.source_model_id,
                        imported_model.logical_name,
                    )
                )
            return


def apply_renamed_model_ids(
    providers: dict[str, Any],
    renamed_model_ids: list[tuple[str, str, str]],
) -> None:
    """Apply queued OpenClaw model-id renames after reference replacement."""

    for provider_key, source_model_id, logical_name in renamed_model_ids:
        provider_config = providers.get(provider_key)
        if not isinstance(provider_config, dict):
            continue
        models = provider_config.get("models")
        if not isinstance(models, list):
            continue
        for model in models:
            if (
                isinstance(model, dict)
                and str(model.get("id", "") or "").strip() == source_model_id
            ):
                model["id"] = logical_name
                break



import importlib
import sys
from pathlib import Path

import pytest

PROJECT_ROOT = Path(__file__).resolve().parents[1]
if str(PROJECT_ROOT) not in sys.path:
    sys.path.insert(0, str(PROJECT_ROOT))

consts = importlib.import_module("cli.consts")
runtime_topology = importlib.import_module("cli.runtime_topology")

RUNTIME_TOPOLOGY_ENV = consts.RUNTIME_TOPOLOGY_ENV
RUNTIME_TOPOLOGY_LEGACY = consts.RUNTIME_TOPOLOGY_LEGACY
RUNTIME_TOPOLOGY_SPLIT = consts.RUNTIME_TOPOLOGY_SPLIT
resolve_runtime_topology = runtime_topology.resolve_runtime_topology
split_runtime_enabled = runtime_topology.split_runtime_enabled


def test_runtime_topology_defaults_to_split(monkeypatch):
    monkeypatch.delenv(RUNTIME_TOPOLOGY_ENV, raising=False)

    assert resolve_runtime_topology() == RUNTIME_TOPOLOGY_SPLIT
    assert split_runtime_enabled() is True


def test_runtime_topology_accepts_split_env(monkeypatch):
    monkeypatch.setenv(RUNTIME_TOPOLOGY_ENV, RUNTIME_TOPOLOGY_SPLIT)

    assert resolve_runtime_topology() == RUNTIME_TOPOLOGY_SPLIT
    assert split_runtime_enabled() is True


def test_runtime_topology_explicit_arg_overrides_env(monkeypatch):
    monkeypatch.setenv(RUNTIME_TOPOLOGY_ENV, RUNTIME_TOPOLOGY_LEGACY)

    assert resolve_runtime_topology(RUNTIME_TOPOLOGY_SPLIT) == RUNTIME_TOPOLOGY_SPLIT
    assert split_runtime_enabled(RUNTIME_TOPOLOGY_SPLIT) is True


def test_runtime_topology_rejects_invalid_values(monkeypatch):
    monkeypatch.setenv(RUNTIME_TOPOLOGY_ENV, "sidecar")

    with pytest.raises(ValueError, match="Invalid runtime topology"):
        resolve_runtime_topology()



import json
import sys
from pathlib import Path

import pytest

PROJECT_ROOT = Path(__file__).resolve().parents[1]
if str(PROJECT_ROOT) not in sys.path:
    sys.path.insert(0, str(PROJECT_ROOT))

from cli import docker_runtime  # noqa: E402


@pytest.fixture(autouse=True)
def clear_runtime_detection_cache():
    docker_runtime._detect_container_runtime.cache_clear()
    yield
    docker_runtime._detect_container_runtime.cache_clear()


def test_detect_container_runtime_accepts_real_docker(monkeypatch):
    monkeypatch.delenv("CONTAINER_RUNTIME", raising=False)
    monkeypatch.setattr(
        docker_runtime.shutil,
        "which",
        lambda name: "/usr/local/bin/docker" if name == "docker" else None,
    )
    monkeypatch.setattr(docker_runtime.os.path, "realpath", lambda path: path)

    class Result:
        def __init__(self, stdout, stderr="", returncode=0):
            self.stdout = stdout
            self.stderr = stderr
            self.returncode = returncode

    def fake_run(cmd, *args, **kwargs):
        if cmd[1] == "--version":
            return Result("Docker version 27.1.1, build deadbeef")
        if cmd[1] == "version":
            return Result(
                json.dumps(
                    {
                        "Client": {"Context": "default"},
                        "Server": {
                            "Components": [{"Name": "Engine", "Version": "27.1.1"}]
                        },
                    }
                )
            )
        raise AssertionError(f"unexpected command: {cmd}")

    monkeypatch.setattr(docker_runtime.subprocess, "run", fake_run)

    assert docker_runtime.get_container_runtime() == "docker"


def test_detect_container_runtime_rejects_podman_env_override(monkeypatch):
    monkeypatch.setenv("CONTAINER_RUNTIME", "podman")

    with pytest.raises(SystemExit):
        docker_runtime.get_container_runtime()


def test_detect_container_runtime_rejects_podman_docker_symlink(monkeypatch):
    monkeypatch.delenv("CONTAINER_RUNTIME", raising=False)
    monkeypatch.setattr(
        docker_runtime.shutil,
        "which",
        lambda name: "/usr/local/bin/docker" if name == "docker" else None,
    )
    monkeypatch.setattr(
        docker_runtime.os.path,
        "realpath",
        lambda path: (
            "/opt/homebrew/bin/podman" if path == "/usr/local/bin/docker" else path
        ),
    )

    with pytest.raises(SystemExit):
        docker_runtime.get_container_runtime()


def test_detect_container_runtime_rejects_when_only_podman_exists(monkeypatch):
    monkeypatch.delenv("CONTAINER_RUNTIME", raising=False)
    monkeypatch.setattr(
        docker_runtime.shutil,
        "which",
        lambda name: "/usr/bin/podman" if name == "podman" else None,
    )

    with pytest.raises(SystemExit):
        docker_runtime.get_container_runtime()


def test_detect_container_runtime_rejects_podman_docker_daemon(monkeypatch):
    monkeypatch.delenv("CONTAINER_RUNTIME", raising=False)
    monkeypatch.setattr(
        docker_runtime.shutil,
        "which",
        lambda name: "/usr/local/bin/docker" if name == "docker" else None,
    )
    monkeypatch.setattr(docker_runtime.os.path, "realpath", lambda path: path)

    class Result:
        def __init__(self, stdout, stderr="", returncode=0):
            self.stdout = stdout
            self.stderr = stderr
            self.returncode = returncode

    def fake_run(cmd, *args, **kwargs):
        if cmd[1] == "--version":
            return Result("Docker version 29.3.0, build deadbeef")
        if cmd[1] == "version":
            return Result(
                json.dumps(
                    {
                        "Client": {"Context": "default"},
                        "Server": {
                            "Components": [
                                {"Name": "Podman Engine", "Version": "5.4.2"}
                            ]
                        },
                    }
                )
            )
        raise AssertionError(f"unexpected command: {cmd}")

    monkeypatch.setattr(docker_runtime.subprocess, "run", fake_run)

    with pytest.raises(SystemExit):
        docker_runtime.get_container_runtime()


def test_detect_container_runtime_rejects_when_docker_daemon_unavailable(monkeypatch):
    monkeypatch.delenv("CONTAINER_RUNTIME", raising=False)
    monkeypatch.setattr(
        docker_runtime.shutil,
        "which",
        lambda name: "/usr/local/bin/docker" if name == "docker" else None,
    )
    monkeypatch.setattr(docker_runtime.os.path, "realpath", lambda path: path)

    class Result:
        def __init__(self, stdout, stderr="", returncode=0):
            self.stdout = stdout
            self.stderr = stderr
            self.returncode = returncode

    def fake_run(cmd, *args, **kwargs):
        if cmd[1] == "--version":
            return Result("Docker version 29.3.0, build deadbeef")
        if cmd[1] == "version":
            return Result(
                "",
                "Cannot connect to the Docker daemon at unix:///Users/test/.docker/run/docker.sock",
                1,
            )
        raise AssertionError(f"unexpected command: {cmd}")

    monkeypatch.setattr(docker_runtime.subprocess, "run", fake_run)

    with pytest.raises(SystemExit):
        docker_runtime.get_container_runtime()


def test_resolve_docker_cli_path_rejects_podman_version_output(monkeypatch):
    monkeypatch.setattr(docker_runtime.os.path, "realpath", lambda path: path)

    class Result:
        stdout = "Emulate Docker CLI using podman. podman version 5.0.0"
        stderr = ""

    monkeypatch.setattr(
        docker_runtime.subprocess, "run", lambda *args, **kwargs: Result()
    )

    with pytest.raises(SystemExit):
        docker_runtime.resolve_docker_cli_path("/usr/local/bin/docker")



import importlib
import sys
from pathlib import Path

PROJECT_ROOT = Path(__file__).resolve().parents[1]
if str(PROJECT_ROOT) not in sys.path:
    sys.path.insert(0, str(PROJECT_ROOT))

core = importlib.import_module("cli.core")
runtime_stack = importlib.import_module("cli.runtime_stack")


def test_check_envoy_status_uses_ready_probe_when_available(monkeypatch):
    stack_layout = runtime_stack.resolve_runtime_stack()
    captured = []

    def fake_exec(container_name, command):
        captured.append((container_name, command))
        return (0, "200", "")

    monkeypatch.setattr(core, "docker_exec", fake_exec)
    monkeypatch.setattr(core, "docker_container_status", lambda _name: "running")

    assert core._check_envoy_status(stack_layout.envoy_container_name, stack_layout)
    assert captured == [
        (
            stack_layout.envoy_container_name,
            [
                "curl",
                "-f",
                "-s",
                "-o",
                "/dev/null",
                "-w",
                "%{http_code}",
                "http://localhost:9901/ready",
            ],
        )
    ]


def test_check_envoy_status_falls_back_to_envoy_validate(monkeypatch):
    stack_layout = runtime_stack.resolve_runtime_stack()
    captured = []
    responses = iter(((127, "", "curl missing"), (0, "", "")))

    def fake_exec(container_name, command):
        captured.append((container_name, command))
        return next(responses)

    monkeypatch.setattr(core, "docker_exec", fake_exec)
    monkeypatch.setattr(core, "docker_container_status", lambda _name: "running")

    assert core._check_envoy_status(stack_layout.envoy_container_name, stack_layout)
    assert captured == [
        (
            stack_layout.envoy_container_name,
            [
                "curl",
                "-f",
                "-s",
                "-o",
                "/dev/null",
                "-w",
                "%{http_code}",
                "http://localhost:9901/ready",
            ],
        ),
        (
            stack_layout.envoy_container_name,
            [
                "/usr/local/bin/envoy",
                "--mode",
                "validate",
                "-c",
                "/etc/envoy/envoy.yaml",
            ],
        ),
    ]


def test_check_envoy_status_does_not_fallback_for_monolith_container(monkeypatch):
    stack_layout = runtime_stack.resolve_runtime_stack()
    captured = []

    def fake_exec(container_name, command):
        captured.append((container_name, command))
        return (127, "", "curl missing")

    monkeypatch.setattr(core, "docker_exec", fake_exec)
    monkeypatch.setattr(core, "docker_container_status", lambda _name: "running")

    assert core._check_envoy_status(stack_layout.container_name, stack_layout) is False
    assert captured == [
        (
            stack_layout.container_name,
            [
                "curl",
                "-f",
                "-s",
                "-o",
                "/dev/null",
                "-w",
                "%{http_code}",
                "http://localhost:9901/ready",
            ],
        )
    ]


def test_show_status_reports_not_running_when_docker_daemon_unreachable(monkeypatch):
    messages = []
    stack_layout = runtime_stack.resolve_runtime_stack()

    monkeypatch.setattr(core, "resolve_runtime_stack", lambda: stack_layout)
    monkeypatch.setattr(
        core,
        "_runtime_stack_status",
        lambda _stack_layout: (_ for _ in ()).throw(SystemExit(1)),
    )
    monkeypatch.setattr(core.log, "info", messages.append)

    core.show_status()

    assert "Status: Not running" in messages
    assert any("Docker daemon is not reachable" in message for message in messages)
    assert "Start with: vllm-sr serve" in messages



"""Derived names and host ports for the local vLLM-SR runtime stack."""

from __future__ import annotations

import os
import re
from dataclasses import dataclass

from cli.consts import (
    DEFAULT_API_PORT,
    DEFAULT_DASHBOARD_PORT,
    DEFAULT_FLEET_SIM_PORT,
    DEFAULT_METRICS_PORT,
    DEFAULT_ROUTER_PORT,
    DEFAULT_STACK_NAME,
    VLLM_SR_DOCKER_NAME,
    VLLM_SR_SIM_DOCKER_NAME,
)

STACK_NAME_ENV = "VLLM_SR_STACK_NAME"
PORT_OFFSET_ENV = "VLLM_SR_PORT_OFFSET"

DEFAULT_JAEGER_OTLP_PORT = 4318
DEFAULT_JAEGER_UI_PORT = 16686
DEFAULT_PROMETHEUS_PORT = 9090
DEFAULT_GRAFANA_PORT = 3000
DEFAULT_REDIS_PORT = 6379
DEFAULT_POSTGRES_PORT = 5432

STACK_NAME_PATTERN = re.compile(r"[^A-Za-z0-9_.-]+")


@dataclass(frozen=True)
class RuntimeStackLayout:
    stack_name: str
    port_offset: int
    container_name: str
    router_container_name: str
    envoy_container_name: str
    dashboard_container_name: str
    fleet_sim_container_name: str
    network_name: str
    jaeger_container_name: str
    prometheus_container_name: str
    grafana_container_name: str
    redis_container_name: str
    postgres_container_name: str
    router_port: int
    metrics_port: int
    dashboard_port: int
    api_port: int
    fleet_sim_port: int
    jaeger_otlp_port: int
    jaeger_ui_port: int
    prometheus_port: int
    grafana_port: int
    redis_port: int
    postgres_port: int

    @property
    def dashboard_url(self) -> str:
        return f"http://localhost:{self.dashboard_port}"

    @property
    def dashboard_service_url(self) -> str:
        return f"http://{self.dashboard_container_name}:8700"

    @property
    def metrics_url(self) -> str:
        return f"http://localhost:{self.metrics_port}/metrics"

    @property
    def router_api_service_url(self) -> str:
        return f"http://{self.router_container_name}:8080"

    @property
    def router_metrics_service_url(self) -> str:
        return f"http://{self.router_container_name}:9190/metrics"

    @property
    def envoy_admin_service_url(self) -> str:
        return f"http://{self.envoy_container_name}:9901"

    def envoy_listener_service_url(self, listener_port: int) -> str:
        return f"http://{self.envoy_container_name}:{listener_port}"

    @property
    def jaeger_ui_url(self) -> str:
        return f"http://localhost:{self.jaeger_ui_port}"

    @property
    def grafana_url(self) -> str:
        return f"http://localhost:{self.grafana_port}"

    @property
    def prometheus_url(self) -> str:
        return f"http://localhost:{self.prometheus_port}"

    @property
    def fleet_sim_url(self) -> str:
        return f"http://localhost:{self.fleet_sim_port}"

    @property
    def jaeger_service_url(self) -> str:
        return f"http://{self.jaeger_container_name}:16686"

    @property
    def prometheus_service_url(self) -> str:
        return f"http://{self.prometheus_container_name}:9090"

    @property
    def grafana_service_url(self) -> str:
        return f"http://{self.grafana_container_name}:3000"

    @property
    def fleet_sim_service_url(self) -> str:
        return f"http://{self.fleet_sim_container_name}:8000"

    @property
    def otlp_service_endpoint(self) -> str:
        return f"http://{self.jaeger_container_name}:4317"

    @property
    def redis_url(self) -> str:
        return f"localhost:{self.redis_port}"

    @property
    def postgres_url(self) -> str:
        return f"localhost:{self.postgres_port}"

    @property
    def storage_container_names(self) -> tuple[str, ...]:
        return (self.redis_container_name, self.postgres_container_name)

    @property
    def runtime_container_names(self) -> tuple[str, ...]:
        return tuple(
            dict.fromkeys(
                (
                    self.container_name,
                    self.router_container_name,
                    self.envoy_container_name,
                    self.dashboard_container_name,
                )
            )
        )

    def service_container_name(self, service: str) -> str:
        if service == "router":
            return self.router_container_name
        if service == "envoy":
            return self.envoy_container_name
        if service == "dashboard":
            return self.dashboard_container_name
        if service == "simulator":
            return self.fleet_sim_container_name
        raise KeyError(f"unknown runtime service: {service}")


def resolve_runtime_stack(
    *, stack_name: str | None = None, port_offset: int | None = None
) -> RuntimeStackLayout:
    resolved_stack_name = normalize_stack_name(
        stack_name if stack_name is not None else os.getenv(STACK_NAME_ENV)
    )
    resolved_port_offset = normalize_port_offset(
        port_offset if port_offset is not None else os.getenv(PORT_OFFSET_ENV)
    )

    if resolved_stack_name == DEFAULT_STACK_NAME:
        container_name = VLLM_SR_DOCKER_NAME
        router_container_name = "vllm-sr-router-container"
        envoy_container_name = "vllm-sr-envoy-container"
        dashboard_container_name = "vllm-sr-dashboard-container"
        fleet_sim_container_name = VLLM_SR_SIM_DOCKER_NAME
        network_name = f"{DEFAULT_STACK_NAME}-network"
        jaeger_container_name = f"{DEFAULT_STACK_NAME}-jaeger"
        prometheus_container_name = f"{DEFAULT_STACK_NAME}-prometheus"
        grafana_container_name = f"{DEFAULT_STACK_NAME}-grafana"
        redis_container_name = f"{DEFAULT_STACK_NAME}-redis"
        postgres_container_name = f"{DEFAULT_STACK_NAME}-postgres"
    else:
        container_name = f"{resolved_stack_name}-vllm-sr-container"
        router_container_name = f"{resolved_stack_name}-vllm-sr-router-container"
        envoy_container_name = f"{resolved_stack_name}-vllm-sr-envoy-container"
        dashboard_container_name = f"{resolved_stack_name}-vllm-sr-dashboard-container"
        fleet_sim_container_name = f"{resolved_stack_name}-vllm-sr-sim"
        network_name = f"{resolved_stack_name}-vllm-sr-network"
        jaeger_container_name = f"{resolved_stack_name}-vllm-sr-jaeger"
        prometheus_container_name = f"{resolved_stack_name}-vllm-sr-prometheus"
        grafana_container_name = f"{resolved_stack_name}-vllm-sr-grafana"
        redis_container_name = f"{resolved_stack_name}-vllm-sr-redis"
        postgres_container_name = f"{resolved_stack_name}-vllm-sr-postgres"

    return RuntimeStackLayout(
        stack_name=resolved_stack_name,
        port_offset=resolved_port_offset,
        container_name=container_name,
        router_container_name=router_container_name,
        envoy_container_name=envoy_container_name,
        dashboard_container_name=dashboard_container_name,
        fleet_sim_container_name=fleet_sim_container_name,
        network_name=network_name,
        jaeger_container_name=jaeger_container_name,
        prometheus_container_name=prometheus_container_name,
        grafana_container_name=grafana_container_name,
        redis_container_name=redis_container_name,
        postgres_container_name=postgres_container_name,
        router_port=DEFAULT_ROUTER_PORT + resolved_port_offset,
        metrics_port=DEFAULT_METRICS_PORT + resolved_port_offset,
        dashboard_port=DEFAULT_DASHBOARD_PORT + resolved_port_offset,
        api_port=DEFAULT_API_PORT + resolved_port_offset,
        fleet_sim_port=DEFAULT_FLEET_SIM_PORT + resolved_port_offset,
        jaeger_otlp_port=DEFAULT_JAEGER_OTLP_PORT + resolved_port_offset,
        jaeger_ui_port=DEFAULT_JAEGER_UI_PORT + resolved_port_offset,
        prometheus_port=DEFAULT_PROMETHEUS_PORT + resolved_port_offset,
        grafana_port=DEFAULT_GRAFANA_PORT + resolved_port_offset,
        redis_port=DEFAULT_REDIS_PORT + resolved_port_offset,
        postgres_port=DEFAULT_POSTGRES_PORT + resolved_port_offset,
    )


def normalize_stack_name(raw_value: str | None) -> str:
    if raw_value is None:
        return DEFAULT_STACK_NAME
    cleaned = STACK_NAME_PATTERN.sub("-", raw_value.strip()).strip("-")
    if not cleaned:
        return DEFAULT_STACK_NAME
    return cleaned


def normalize_port_offset(raw_value: str | int | None) -> int:
    if raw_value in (None, ""):
        return 0
    offset = int(raw_value)
    if offset < 0:
        raise ValueError(f"{PORT_OFFSET_ENV} must be >= 0, got {offset}")
    return offset
